"""
Calculation of the optimal power flow.
"""

import logging

import numpy as np

from hynet.types_ import hynet_float_, SolverType, SolverStatus
from hynet.solver import AVAILABLE_SOLVERS
from hynet.data.connection import DBConnection
from hynet.data.interface import load_scenario
from hynet.scenario.representation import Scenario
from hynet.model.steady_state import SystemModel
from hynet.qcqp.solver import SolverInterface
from hynet.qcqp.problem import QCQPPoint, QCQP
from hynet.utilities.base import Timer
from hynet.reduction.copper_plate import reduce_to_copper_plate
import hynet.config as config

_log = logging.getLogger(__name__)


def calc_opf(data, scenario_id=0, solver=None, solver_type=None):
    """
    Calculate the optimal power flow.

    This function formulates and solves the optimal power flow (OPF) problem.
    The solver or solver type may be specified explicitly, otherwise an
    appropriate solver is selected automatically.

    **Custom OPF Formulations:** For a customization of the standard OPF
    formulation, please refer to the docstring of ``SystemModel``.

    **Copper Plate Models:** *hynet* supports the simulation of "copper plate
    models", i.e., when the grid model is neglected all injectors and
    loads are connected to a single bus. As such models exhibit the *hybrid
    architecture*, all solver types (QCQP, SDR, and SOCR) are applicable. The
    grid model of a scenario can be reduced to a copper plate using
    ``reduce_to_copper_plate``.

    Parameters
    ----------
    data : DBConnection or Scenario or SystemModel
        Connection to a *hynet* grid database, a ``Scenario`` object, or a
        ``SystemModel`` object.
    scenario_id : .hynet_id_, optional
        Identifier of the scenario. This argument is ignored if ``data`` is a
        ``Scenario`` or ``SystemModel`` object.
    solver : SolverInterface, optional
        Solver for the QCQP problem; the default selects an appropriate
        solver of those available.
    solver_type : SolverType, optional
        Optional solver type for the QCQP problem, i.e., if passed it restricts
        the automatic solver selection to this type. It is ignored if
        ``solver`` is not ``None``.

    Returns
    -------
    result : OPFResult
        Optimal power flow solution.

    See Also
    --------
    hynet.scenario.representation.Scenario
    hynet.opf.result.OPFResult
    hynet.types_.SolverType
    hynet.model.steady_state.SystemModel
    hynet.reduction.copper_plate.reduce_to_copper_plate
    """
    timer = Timer()
    if isinstance(data, SystemModel):
        model = data
    elif isinstance(data, Scenario):
        model = SystemModel(data)
    elif isinstance(data, DBConnection):
        model = SystemModel(load_scenario(data, scenario_id))
    else:
        raise ValueError(("The argument 'data' must be a database file name, "
                          "a Scenario object, or a SystemModel object."))

    if solver is None:
        solver = select_solver(model, solver_type)()
    elif not isinstance(solver, SolverInterface):
        raise ValueError("The solver must be a SolverInterface-derived object.")

    if solver.type == SolverType.SDR and model.dim_v > 500:
        _log.warning("Solving the SDR for this large system potentially "
                     "requires an excessive amount of time.")

    _log.debug("Calculate OPF ~ Loading, verification, and solver selection "
               "({:.3f} sec.)".format(timer.interval()))
    timer.reset()

    qcqp = model.get_opf_problem()

    if solver.type == SolverType.QCQP and model.dim_v > len(model.islands):
        # Support the solver with an appropriate initial point
        if config.OPF['copper_plate_init']:
            qcqp.initial_point = get_copper_plate_initial_point(model, qcqp)

    _log.debug("Calculate OPF ~ QCQP creation ({:.3f} sec.)"
               .format(timer.interval()))

    result = solver.solve(qcqp)
    return model.create_opf_result(result, total_time=timer.total())


def get_relaxation_initial_point(model, qcqp, solver_type):
    """
    Return a relaxation-based initial point for the given OPF QCQP.

    This method returns an initial point for the solution of the OPF QCQP that
    corresponds to a relaxation of the OPF. Especially the second-order cone
    relaxation (SOCR solvers) is typically fast to compute and may be suitable.
    However, in our experience with meshed grids, only in very rare cases the
    SOCR-based initialization turned out to be superior to a copper plate based
    initialization.

    Parameters
    ----------
    model : SystemModel
        Steady-state system model, for which the OPF shall be calculated.
    qcqp : QCQP
        QCQP for the OPF of the given model, for which the initial point
        shall be determined.
    solver_type : SolverType
        Solver type that shall be used to compute the initial point.

    Returns
    -------
    initial_point : QCQPPoint or None
        Relaxation-based initial point for the given OPF QCQP. In case the
        solution failed or no appropriate solver was found, None is returned.
    """
    timer = Timer()
    initial_point = None
    try:
        solver = select_solver(model, solver_type)()
        result = solver.solve(qcqp)
        if result.solver_status != SolverStatus.SOLVED:
            raise RuntimeError("OPF failed with status '"
                               + str(result.solver_status) + "'.")
        initial_point = result.optimizer.scale(qcqp.normalization)
    except RuntimeError as exception:
        _log.warning("{0}-based initialization failed: {1}"
                     .format(solver_type.value, str(exception)))
    _log.debug("{0}-based initial point calculation ({1:.3f} sec.)"
               .format(solver_type.value, timer.total()))
    return initial_point


def get_copper_plate_initial_point(model, qcqp):
    """
    Return a copper plate based initial point for the given OPF QCQP.

    This method returns an initial point for the solution of the OPF QCQP that
    includes the optimal dispatch of the copper plate reduction of the model.
    The copper plate solution is typically fast to compute and can reduce the
    number of iterations required to solve the nonconvex QCQP, i.e., it can
    improve the overall performance when solving the nonconvex problem.

    Parameters
    ----------
    model : SystemModel
        Steady-state system model, for which the OPF shall be calculated.
    qcqp : QCQP
        QCQP for the OPF of the given model, for which the initial point
        shall be determined.

    Returns
    -------
    initial_point : QCQPPoint or None
        Copper plate model based initial point for the given OPF QCQP. In case
        the solution failed or no SOCR solver was found, None is returned.
    """
    timer = Timer()
    initial_point = None
    try:
        cp = SystemModel(reduce_to_copper_plate(model.scenario),
                         verify_scenario=False)
        cp_solver = select_solver(cp, SolverType.SOCR)()
        cp_result = cp_solver.solve(cp.get_opf_problem())
        if cp_result.solver_status != SolverStatus.SOLVED:
            raise RuntimeError("OPF failed with status '"
                               + str(cp_result.solver_status) + "'.")
        initial_point = QCQPPoint(v=0.35 * qcqp.lb.v + 0.65 * qcqp.ub.v,
                                  f=np.zeros(qcqp.dim_f, dtype=hynet_float_),
                                  s=cp_result.optimizer.s * qcqp.normalization.s,
                                  z=np.zeros(qcqp.dim_z, dtype=hynet_float_))
    except RuntimeError as exception:
        _log.warning("Copper plate initialization failed: " + str(exception))
    _log.debug("Copper plate initial point calculation ({:.3f} sec.)"
               .format(timer.total()))
    return initial_point


def select_solver(model, solver_type=None):
    """
    Return the most appropriate solver of those available for the given conditions.

    Parameters
    ----------
    model : SystemModel
        Steady state model for which the OPF shall be calculated.
    solver_type : SolverType, optional
        Specification of the solver type. By default, the most appropriate
        among all available solver types is selected.

    Returns
    -------
    solver : SolverInterface
        Selected solver interface *class*.

    Raises
    ------
    RuntimeError
        In case no appropriate solver was found.
    """
    if not AVAILABLE_SOLVERS:
        raise RuntimeError("No supported solvers were found.")

    qcqp_solvers = [x for x in AVAILABLE_SOLVERS if x().type == SolverType.QCQP]
    sdr_solvers = [x for x in AVAILABLE_SOLVERS if x().type == SolverType.SDR]
    socr_solvers = [x for x in AVAILABLE_SOLVERS if x().type == SolverType.SOCR]

    if solver_type is not None:
        if solver_type == SolverType.QCQP:
            if qcqp_solvers:
                solver = qcqp_solvers[0]
            else:
                raise RuntimeError("No supported QCQP solvers were found.")
        elif solver_type == SolverType.SDR:
            if sdr_solvers:
                solver = sdr_solvers[0]
            else:
                raise RuntimeError("No supported SDR solvers were found.")
        elif solver_type == SolverType.SOCR:
            if socr_solvers:
                solver = socr_solvers[0]
            else:
                raise RuntimeError("No supported SOCR solvers were found.")
        else:
            raise ValueError("The solver type must be a SolverType member.")
    else:
        if model.is_acyclic:
            # For the hybrid architecture, choose the most efficient solver
            if socr_solvers:
                solver = socr_solvers[0]
            elif sdr_solvers:
                solver = sdr_solvers[0]
            else:
                solver = qcqp_solvers[0]
        else:
            # ...and for conventional grids in the order of "tightness"
            if qcqp_solvers:
                solver = qcqp_solvers[0]
            elif sdr_solvers and model.dim_v <= 300:
                # Only for smaller systems due to computational complexity...
                solver = sdr_solvers[0]
            else:
                solver = socr_solvers[0]
    return solver



#pylint: disable=too-few-public-methods,C0330
"""
Schema and SQLAlchemy declaratives for *hynet* grid databases.
"""

import math

from sqlalchemy import (Column,
                        String,
                        Integer,
                        Boolean,
                        Float,
                        Enum,
                        ForeignKey)
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship

from hynet.types_ import (hynet_eps,
                          BusType,
                          BranchType,
                          InjectorType,
                          EntityType,
                          DBInfoKey)


SCHEMA_VERSION = '1.0'  # Version of the database scheme.

DEFAULT_SHORT_STRING_LENGTH = 1024
DEFAULT_LONG_STRING_LENGTH = 65535

Base = declarative_base()


def _enum_to_value(x):
    return [e.value for e in x]


class DBInfo(Base):
    """Dataset for a *hynet* grid database setting."""

    __tablename__ = 'db_info'

    # Use an unconstrained enum to parse keys from database
    key = Column(Enum(DBInfoKey,
                      native_enum=False,
                      values_callable=_enum_to_value,
                      create_constraint=False,
                      validate_strings=False),
                 primary_key=True, nullable=False)
    value = Column(String(DEFAULT_LONG_STRING_LENGTH), nullable=False)


class DBBus(Base):
    """Dataset for a bus."""

    __tablename__ = 'bus'

    id = Column(Integer, primary_key=True)
    type = Column(Enum(BusType, values_callable=_enum_to_value),
                  nullable=False)
    ref = Column(Boolean, nullable=False)
    base_kv = Column(Float, nullable=False)
    v_min = Column(Float, nullable=False)
    v_max = Column(Float, nullable=False)
    zone = Column(Integer, nullable=True)
    annotation = Column(String(DEFAULT_LONG_STRING_LENGTH), nullable=False)

    def __repr__(self):
        return "{0.id}: {0.type}, {0.annotation}".format(self)

    def __eq__(self, other):
        """Return True if the data is equivalent."""
        if other is None:
            return False
        return (self.id == other.id and
                self.type == other.type and
                self.ref == other.ref and
                self.base_kv == other.base_kv and
                self.v_min == other.v_min and
                self.v_max == other.v_max and
                self.zone == other.zone and
                self.annotation == other.annotation)

    def __hash__(self):
        return self.id


class DBBranch(Base):
    """Dataset for a branch."""

    __tablename__ = 'branch'

    id = Column(Integer, primary_key=True)
    type = Column(Enum(BranchType, values_callable=_enum_to_value),
                  nullable=False)
    src_id = Column(Integer, ForeignKey('bus.id'))
    src = relationship('DBBus', foreign_keys=[src_id])
    dst_id = Column(Integer, ForeignKey('bus.id'))
    dst = relationship('DBBus', foreign_keys=[dst_id])
    r = Column(Float, nullable=False)
    x = Column(Float, nullable=False)
    b_src = Column(Float, nullable=False)
    b_dst = Column(Float, nullable=False)
    ratio_src = Column(Float, nullable=False)
    phase_src = Column(Float, nullable=False)
    ratio_dst = Column(Float, nullable=False)
    phase_dst = Column(Float, nullable=False)
    length = Column(Float, nullable=True)
    rating = Column(Float, nullable=True)
    angle_min = Column(Float, nullable=True)
    angle_max = Column(Float, nullable=True)
    drop_min = Column(Float, nullable=True)
    drop_max = Column(Float, nullable=True)
    annotation = Column(String(DEFAULT_LONG_STRING_LENGTH), nullable=False)

    def __repr__(self):
        return "{0.id}: {0.src_id}, {0.dst_id}, \"{0.annotation}\"" \
            .format(self)

    def __eq__(self, other):
        """Return True if the data is equivalent."""
        if other is None:
            return False
        return (self.id == other.id and
                self.type == other.type and
                self.src_id == other.src_id and
                self.dst_id == other.dst_id and
                self.r == other.r and
                self.x == other.x and
                self.b_src == other.b_src and
                self.b_dst == other.b_dst and
                math.isclose(self.ratio_src, other.ratio_src,
                             rel_tol=hynet_eps * 10) and
                math.isclose(self.phase_src, other.phase_src,
                             rel_tol=hynet_eps * 10) and
                math.isclose(self.ratio_dst, other.ratio_dst,
                             rel_tol=hynet_eps * 10) and
                math.isclose(self.phase_dst, other.phase_dst,
                             rel_tol=hynet_eps * 10) and
                self.length == other.length and
                self.rating == other.rating and
                self.angle_min == other.angle_min and
                self.angle_max == other.angle_max and
                self.drop_min == other.drop_min and
                self.drop_max == other.drop_max and
                self.annotation == other.annotation)

    def __hash__(self):
        return self.id


class DBConverter(Base):
    """Dataset for a converter."""

    __tablename__ = 'converter'

    id = Column(Integer, primary_key=True)
    src_id = Column(Integer, ForeignKey('bus.id'))
    src = relationship('DBBus', foreign_keys=[src_id])
    dst_id = Column(Integer, ForeignKey('bus.id'))
    dst = relationship('DBBus', foreign_keys=[dst_id])
    cap_src_id = Column(Integer, ForeignKey('capability_region.id'))
    cap_src = relationship('DBCapabilityRegion', foreign_keys=[cap_src_id])
    cap_dst_id = Column(Integer, ForeignKey('capability_region.id'))
    cap_dst = relationship('DBCapabilityRegion', foreign_keys=[cap_dst_id])
    loss_fwd = Column(Float, nullable=False)
    loss_bwd = Column(Float, nullable=False)
    loss_fix = Column(Float, nullable=False)
    annotation = Column(String(DEFAULT_LONG_STRING_LENGTH), nullable=False)

    def __eq__(self, other):
        """
        Return True if the data *excl. the capability regions* is equivalent.

        **This is a customized equality operation and designed for the use
        during the saving of scenarios to a database. Apply with caution.**
        """
        if other is None:
            return False
        return (self.id == other.id and
                self.src_id == other.src_id and
                self.dst_id == other.dst_id and
                # self.cap_src_id == other.cap_src_id and
                # self.cap_dst_id == other.cap_dst_id and
                self.loss_fwd == other.loss_fwd and
                self.loss_bwd == other.loss_bwd and
                self.loss_fix == other.loss_fix and
                self.annotation == other.annotation)

    def __hash__(self):
        return self.id


class DBCapabilityRegion(Base):
    """Dataset for a capability region."""

    __tablename__ = 'capability_region'

    id = Column(Integer, primary_key=True)
    p_min = Column(Float, nullable=False)
    p_max = Column(Float, nullable=False)
    q_min = Column(Float, nullable=False)
    q_max = Column(Float, nullable=False)
    lt_ofs = Column(Float, nullable=True)
    lt_slp = Column(Float, nullable=True)
    rt_ofs = Column(Float, nullable=True)
    rt_slp = Column(Float, nullable=True)
    lb_ofs = Column(Float, nullable=True)
    lb_slp = Column(Float, nullable=True)
    rb_ofs = Column(Float, nullable=True)
    rb_slp = Column(Float, nullable=True)
    annotation = Column(String(DEFAULT_LONG_STRING_LENGTH), nullable=False)

    def __eq__(self, other):
        """
        Return True if the data *excl. the ID, box, and annotation* is equivalent.

        **This is a customized equality operation and designed for the use
        during the saving of scenarios to a database. Apply with caution.**
        """
        if other is None:
            return False
        return (# self.id == other.id and
                # self.p_min == other.p_min and
                # self.p_max == other.p_max and
                # self.q_min == other.q_min and
                # self.q_max == other.q_max and
                self.lt_ofs == other.lt_ofs and
                self.lt_slp == other.lt_slp and
                self.rt_ofs == other.rt_ofs and
                self.rt_slp == other.rt_slp and
                self.lb_ofs == other.lb_ofs and
                self.lb_slp == other.lb_slp and
                self.rb_ofs == other.rb_ofs and
                self.rb_slp == other.rb_slp)
                # self.annotation == other.annotation)

    def __hash__(self):
        return self.id


class DBShunt(Base):
    """Dataset for a shunt."""

    __tablename__ = 'shunt'

    id = Column(Integer, primary_key=True)
    bus_id = Column(Integer, ForeignKey('bus.id'))
    bus = relationship('DBBus', foreign_keys=[bus_id])
    p = Column(Float, nullable=False)
    q = Column(Float, nullable=False)
    annotation = Column(String(DEFAULT_LONG_STRING_LENGTH), nullable=False)

    def __eq__(self, other):
        """
        Return True if the data *excl. the ID and annotation* is equivalent.

        **This is a customized equality operation and designed for the use
        during the saving of scenarios to a database. Apply with caution.**
        """
        if other is None:
            return False
        return (# self.id == other.id and
                self.bus_id == other.bus_id and
                self.p == other.p and
                self.q == other.q)
                # self.annotation == other.annotation)

    def __hash__(self):
        return self.id


class DBInjector(Base):
    """Dataset for an injector."""

    __tablename__ = 'injector'

    id = Column(Integer, primary_key=True)
    type = Column(Enum(InjectorType, values_callable=_enum_to_value),
                  nullable=False)
    bus_id = Column(Integer, ForeignKey('bus.id'))
    bus = relationship('DBBus', foreign_keys=[bus_id])
    cap_id = Column(Integer, ForeignKey('capability_region.id'))
    cap = relationship('DBCapabilityRegion', foreign_keys=[cap_id])
    cost_p_id = Column(Integer, ForeignKey('sample_point.id'))
    cost_p = relationship('DBSamplePoint', foreign_keys=[cost_p_id],
                          uselist=True)
    cost_q_id = Column(Integer, ForeignKey('sample_point.id'))
    cost_q = relationship('DBSamplePoint', foreign_keys=[cost_q_id],
                          uselist=True)
    cost_start = Column(Float, nullable=False)
    cost_stop = Column(Float, nullable=False)
    ramp_up = Column(Float, nullable=True)
    ramp_down = Column(Float, nullable=True)
    min_up = Column(Float, nullable=True)
    min_down = Column(Float, nullable=True)
    energy_min = Column(Float, nullable=True)
    energy_max = Column(Float, nullable=True)
    annotation = Column(String(DEFAULT_LONG_STRING_LENGTH), nullable=False)

    def __eq__(self, other):
        """
        Return True if the data *excl. the cap. region and cost IDs* is equivalent.

        **This is a customized equality operation and designed for the use
        during the saving of scenarios to a database. Apply with caution.**
        """
        if other is None:
            return False
        return (self.id == other.id and
                self.type == other.type and
                self.bus_id == other.bus_id and
                # self.cap_id == other.cap_id and
                # self.cost_p_id == other.cost_p_id and
                # self.cost_q_id == other.cost_q_id and
                self.cost_start == other.cost_start and
                self.cost_stop == other.cost_stop and
                self.ramp_up == other.ramp_up and
                self.ramp_down == other.ramp_down and
                self.min_up == other.min_up and
                self.min_down == other.min_down and
                self.energy_min == other.energy_min and
                self.energy_max == other.energy_max and
                self.annotation == other.annotation)

    def __hash__(self):
        return self.id


class DBSamplePoint(Base):
    """Dataset for a cost function sample point."""

    __tablename__ = 'sample_point'

    id = Column(Integer, primary_key=True)
    x = Column(Float, primary_key=True)
    y = Column(Float, nullable=False)


class DBScenario(Base):
    """Dataset for a scenario."""

    __tablename__ = 'scenario'

    id = Column(Integer, primary_key=True)
    time = Column(Float, nullable=False)
    name = Column(String(DEFAULT_LONG_STRING_LENGTH), nullable=False)
    loss_price = Column(Float, nullable=False)
    annotation = Column(String(DEFAULT_LONG_STRING_LENGTH), nullable=False)


class DBScenarioLoad(Base):
    """Dataset for a load in a scenario."""

    __tablename__ = 'scenario_load'

    scenario_id = Column(Integer, ForeignKey('scenario.id'), primary_key=True)
    scenario = relationship('DBScenario', foreign_keys=[scenario_id])
    bus_id = Column(Integer, ForeignKey('bus.id'), primary_key=True)
    bus = relationship('DBBus', foreign_keys=[bus_id])
    p = Column(Float, nullable=False)
    q = Column(Float, nullable=False)


class DBScenarioInjector(Base):
    """Dataset for the injector adaptation in a scenario."""

    __tablename__ = 'scenario_injector'

    scenario_id = Column(Integer, ForeignKey('scenario.id'), primary_key=True)
    scenario = relationship('DBScenario', foreign_keys=[scenario_id])
    injector_id = Column(Integer, ForeignKey('injector.id'), primary_key=True)
    injector = relationship('DBInjector', foreign_keys=[injector_id])
    p_min = Column(Float, nullable=False)
    p_max = Column(Float, nullable=False)
    q_min = Column(Float, nullable=False)
    q_max = Column(Float, nullable=False)
    cost_scaling = Column(Float, nullable=False)


class DBScenarioInactivity(Base):
    """Dataset for the inactivity specification in a scenario."""

    __tablename__ = 'scenario_inactivity'

    scenario_id = Column(Integer, ForeignKey('scenario.id'), primary_key=True)
    scenario = relationship('DBScenario', foreign_keys=[scenario_id])
    entity_type = Column(Enum(EntityType, values_callable=_enum_to_value),
                         primary_key=True)
    entity_id = Column(Integer, primary_key=True)



"""
Visualization functionality for optimal power flow results.
"""

import logging

import numpy as np
import matplotlib.pyplot as plt
from matplotlib.ticker import MaxNLocator

from hynet.types_ import SolverType

_log = logging.getLogger(__name__)


def _check_data_availability(result):
    """Check availability of OPF result data."""
    if result.empty:
        raise RuntimeError("The OPF result is empty.")
    if not result.is_physical:
        if result.solver.type == SolverType.QCQP:
            remark = "solver issue"
        else:
            remark = "inexact relaxation or solver issue"
        _log.warning("The OPF result does not satisfy the power "
                     "balance ({0:s}).".format(remark))


def _set_id_labels(axis, index):
    """
    Relabel the one-based linear ticks with the IDs in ``index``.
    """
    axis.set_xticklabels(
        ['' if not (i.is_integer() and 1 <= i <= len(index))
         else str(index[int(i - 1)]) for i in axis.get_xticks()])


def show_lmp_profile(result, id_label=False):
    """
    Show the LMP profile or power balance dual variables for an OPF result.

    For a consistent appearance in case of nonconsecutive or custom IDs,
    **the buses are numbered consecutively starting from 1 for the
    labeling of the x-axis**. To enforce labeling of the ticks with the
    IDs, set ``id_label=True``.

    Parameters
    ----------
    result : OPFResult
        Optimal power flow result that shall be visualized.
    id_label : bool, optional
        If True, the linear ticks are relabeled with the bus IDs.
    """
    _check_data_availability(result)
    if result.num_buses < 2:
        raise RuntimeError("LMP profile visualization is only available "
                           "for two or more buses.")
    if result.solver.type == SolverType.QCQP:
        # Zero duality gap is not ensured...
        title = "KKT multipliers for {:s} power balance"
    elif not result.is_physical:
        # Relaxation is not exact...
        title = "Dual variables for {:s} power balance"
    else:
        title = "LMP profile for {:s} power"

    fig = plt.figure(figsize=(10, 6))
    ax = fig.add_subplot(2, 1, 1)
    linear_index = 1 + np.arange(result.num_buses)
    ax.plot(linear_index, result.bus['dv_bal_p'].values,
            color='xkcd:sea blue', linestyle='-', linewidth=1)
    ax.margins(x=0)
    ax.xaxis.set_major_locator(MaxNLocator(integer=True))
    if id_label:
        _set_id_labels(ax, result.bus.index)
    ax.set_title(title.format('active'))
    ax.set_ylabel('$/MWh')
    ax.set_xlabel('Bus ' + ('ID' if id_label else 'Index'))
    ax = fig.add_subplot(2, 1, 2)
    ax.plot(linear_index, result.bus['dv_bal_q'].values,
            color='xkcd:sea blue', linestyle='-', linewidth=1)
    ax.margins(x=0)
    ax.xaxis.set_major_locator(MaxNLocator(integer=True))
    if id_label:
        _set_id_labels(ax, result.bus.index)
    ax.set_title(title.format('reactive'))
    ax.set_ylabel('$/Mvarh')
    ax.set_xlabel('Bus ' + ('ID' if id_label else 'Index'))
    fig.tight_layout()
    fig.show()
    return fig


def show_voltage_profile(result, id_label=False):
    """
    Show the voltage profile for an OPF result.

    For a consistent appearance in case of nonconsecutive or custom IDs,
    **the buses are numbered consecutively starting from 1 for the
    labeling of the x-axis**. To enforce labeling of the ticks with the
    IDs, set ``id_label=True``.

    Parameters
    ----------
    result : OPFResult
        Optimal power flow result that shall be visualized.
    id_label : bool, optional
        If True, the linear ticks are relabeled with the bus IDs.
    """
    _check_data_availability(result)
    if result.num_buses < 2:
        raise RuntimeError("Voltage profile visualization is only "
                           "available for two or more buses.")
    fig = plt.figure(figsize=(10, 3))
    ax = fig.add_subplot(1, 1, 1)
    linear_index = 1 + np.arange(result.num_buses)
    v_min = result.scenario.bus.loc[result.bus.index, 'v_min']
    ax.plot(linear_index, v_min.values,
            color='xkcd:light red', linestyle='--', linewidth=0.5)
    v_max = result.scenario.bus.loc[result.bus.index, 'v_max']
    ax.plot(linear_index, v_max.values,
            color='xkcd:light red', linestyle='--', linewidth=0.5)
    ax.plot(linear_index, result.bus['v'].abs().values,
            color='xkcd:sea blue', linestyle='-', linewidth=1)
    ax.margins(x=0)
    ax.xaxis.set_major_locator(MaxNLocator(integer=True))
    if id_label:
        _set_id_labels(ax, result.bus.index)
    ax.set_title("Voltage profile")
    ax.set_ylabel('p.u.')
    ax.set_xlabel('Bus ' + ('ID' if id_label else 'Index'))
    fig.tight_layout()
    fig.show()
    return fig


def show_branch_flow_profile(result, id_label=False):
    """
    Show the branch flow profile for an OPF result.

    For a consistent appearance in case of nonconsecutive or custom IDs,
    **the branches are numbered consecutively starting from 1 for the
    labeling of the x-axis**. To enforce labeling of the ticks with the
    IDs, set ``id_label=True``.

    Parameters
    ----------
    result : OPFResult
        Optimal power flow result that shall be visualized.
    id_label : bool, optional
        If True, the linear ticks are relabeled with the branch IDs.
    """
    _check_data_availability(result)
    if result.num_branches < 2:
        raise RuntimeError("Branch flow profile visualization is only "
                           "available for two or more branches.")
    fig = plt.figure(figsize=(10, 3))
    ax = fig.add_subplot(1, 1, 1)
    linear_index = 1 + np.arange(result.num_branches)
    ax.plot(linear_index, result.branch['effective_rating'].values,
            color='xkcd:light red', linestyle='--', linewidth=0.5)
    ax.plot(linear_index,
            result.branch[['s_src', 's_dst']].abs().max(axis=1).values,
            color='xkcd:sea blue', linestyle='-', linewidth=1)
    ax.margins(x=0)
    ax.xaxis.set_major_locator(MaxNLocator(integer=True))
    if id_label:
        _set_id_labels(ax, result.branch.index)
    ax.set_title("Branch flow profile")
    ax.set_ylabel('MVA for AC / MW for DC')
    ax.set_xlabel('Branch ' + ('ID' if id_label else 'Index'))
    fig.tight_layout()
    fig.show()
    return fig


def show_ampacity_dual_profile(result, id_label=False):
    """
    Show the ampacity dual variable profile for an OPF result.

    For a consistent appearance in case of nonconsecutive or custom IDs,
    **the branches are numbered consecutively starting from 1 for the
    labeling of the x-axis**. To enforce labeling of the ticks with the
    IDs, set ``id_label=True``.

    Parameters
    ----------
    result : OPFResult
        Optimal power flow result that shall be visualized.
    id_label : bool, optional
        If True, the linear ticks are relabeled with the branch IDs.
    """
    _check_data_availability(result)
    if result.num_branches < 2:
        raise RuntimeError("Ampacity dual variable profile visualization "
                           "is only available for two or more branches.")
    dv_ampacity = result.branch[['dv_i_max_src',
                                 'dv_i_max_dst']].sum(axis=1).values
    fig = plt.figure(figsize=(10, 3))
    ax = fig.add_subplot(1, 1, 1)
    linear_index = 1 + np.arange(result.num_branches)
    ax.plot(linear_index, dv_ampacity,
            color='xkcd:sea blue', linestyle='-', linewidth=1)
    ax.margins(x=0)
    ax.xaxis.set_major_locator(MaxNLocator(integer=True))
    if id_label:
        _set_id_labels(ax, result.branch.index)
    ax.set_title("Ampacity dual variable profile")
    ax.set_ylabel('Dual Variable')
    ax.set_xlabel('Branch ' + ('ID' if id_label else 'Index'))
    fig.tight_layout()
    fig.show()
    return fig


def show_converter_flow_profile(result, id_label=False):
    """
    Show the converter active power flow profile for an OPF result.

    For a consistent appearance in case of nonconsecutive or custom IDs,
    **the converters are numbered consecutively starting from 1 for the
    labeling of the x-axis**. To enforce labeling of the ticks with the
    IDs, set ``id_label=True``.

    Parameters
    ----------
    result : OPFResult
        Optimal power flow result that shall be visualized.
    id_label : bool, optional
        If True, the linear ticks are relabeled with the converter IDs.
    """
    _check_data_availability(result)
    if result.num_converters < 2:
        raise RuntimeError("Converter flow profile visualization is only "
                           "available for two or more converters.")
    fig = plt.figure(figsize=(10, 6))
    ax = fig.add_subplot(2, 1, 1)
    index = result.converter.index
    linear_index = 1 + np.arange(result.num_converters)
    cap_src = result.scenario.converter.loc[index, 'cap_src']
    ax.plot(linear_index, [x.p_min for x in cap_src],
            color='xkcd:light red', linestyle='--', linewidth=0.5)
    ax.plot(linear_index, [x.p_max for x in cap_src],
            color='xkcd:light red', linestyle='--', linewidth=0.5)
    ax.plot(linear_index, result.converter['p_src'].values,
            color='xkcd:sea blue', linestyle='-', linewidth=1)
    ax.margins(x=0)
    ax.xaxis.set_major_locator(MaxNLocator(integer=True))
    if id_label:
        _set_id_labels(ax, index)
    ax.set_title("Converter flow profile at the source terminal")
    ax.set_ylabel('MW')
    ax.set_xlabel('Converter ' + ('ID' if id_label else 'Index'))
    ax = fig.add_subplot(2, 1, 2)
    cap_dst = result.scenario.converter.loc[index, 'cap_dst']
    ax.plot(linear_index, [x.p_min for x in cap_dst],
            color='xkcd:light red', linestyle='--', linewidth=0.5)
    ax.plot(linear_index, [x.p_max for x in cap_dst],
            color='xkcd:light red', linestyle='--', linewidth=0.5)
    ax.plot(linear_index, result.converter['p_dst'].values,
            color='xkcd:sea blue', linestyle='-', linewidth=1)
    ax.margins(x=0)
    ax.xaxis.set_major_locator(MaxNLocator(integer=True))
    if id_label:
        _set_id_labels(ax, index)
    ax.set_title("Converter flow profile at the destination terminal")
    ax.set_ylabel('MW')
    ax.set_xlabel('Converter ' + ('ID' if id_label else 'Index'))
    fig.tight_layout()
    fig.show()
    return fig


def show_dispatch_profile(result, id_label=False):
    """
    Show the active power injector dispatch profile for an OPF result.

    For a consistent appearance in case of nonconsecutive or custom IDs,
    **the injectors are numbered consecutively starting from 1 for the
    labeling of the x-axis**. To enforce labeling of the ticks with the
    IDs, set ``id_label=True``.

    Parameters
    ----------
    result : OPFResult
        Optimal power flow result that shall be visualized.
    id_label : bool, optional
        If True, the linear ticks are relabeled with the injector IDs.
    """
    _check_data_availability(result)
    if result.num_injectors < 2:
        raise RuntimeError("Injector dispatch visualization is only "
                           "available for two or more injectors.")
    fig = plt.figure(figsize=(10, 3))
    ax = fig.add_subplot(1, 1, 1)
    linear_index = 1 + np.arange(result.num_injectors)
    cap = result.scenario.injector.loc[result.injector.index, 'cap']
    ax.plot(linear_index, [x.p_min for x in cap],
            color='xkcd:light red', linestyle='--', linewidth=0.5)
    ax.plot(linear_index, [x.p_max for x in cap],
            color='xkcd:light red', linestyle='--', linewidth=0.5)
    ax.plot(linear_index, result.injector['s'].values.real,
            color='xkcd:sea blue', linestyle='-', linewidth=1)
    ax.margins(x=0)
    ax.xaxis.set_major_locator(MaxNLocator(integer=True))
    if id_label:
        _set_id_labels(ax, result.injector.index)
    ax.set_title("Active power injector dispatch profile")
    ax.set_ylabel('MW')
    ax.set_xlabel('Injector ' + ('ID' if id_label else 'Index'))
    fig.tight_layout()
    fig.show()
    return fig



#pylint: disable=no-member
"""
Numeric types and enumerations in *hynet*.
"""

from enum import Enum

import numpy as np
from scipy.sparse import coo_matrix

hynet_id_ = np.int64
hynet_int_ = np.int64
hynet_float_ = np.float64
hynet_complex_ = np.complex128
hynet_sparse_ = coo_matrix

# REMARK to sparse matrices in *hynet*:
# -------------------------------------
#
# First of all, KEEP THE COORDINATE SPARSITY FORMAT, even if the definition
# suggests that it may be modified ;) For performance reasons, I had to utilize
# its particular data organisation and a modification would break some code.
#
# Anyway, after several rounds of performance optimization the selection of the
# SciPy COO format is well motivated. Besides the fast creation of sparse
# matrices (e.g. in SystemModel) and enabling of efficient data reorganisation
# (e.g. in QCQP), the most profound reason for its choice is memory efficiency.
# For larger grids, thousands of sparse matrices are generated to construct the
# OPF QCQP and, for the originally employed CSR format (which is beneficial for
# slicing, indexing, and arithmetic), this soon leads to QCQP objects of 100MB
# and more. This high memory consumption is a particular issue with parallel
# and distributed processing, where certain objects are pickled. For example,
# during the constraint generation in SystemModel.get_opf_problem, this caused
# Python's multiprocessing package to crash for the 13659-bus grid of the
# PGLib. Switching to the COO format reduced the memory consumption
# considerably, e.g., for scenario 0 of the German 2030 NEP grid with 1524
# buses the pickle of its QCQP object was reduced from 113.7MB to 7.9MB, i.e.,
# by 93%.
#
# Finally, a particularity of COO matrices should be kept in mind, e.g., if a
# a new solver interface is implemented. This format permits duplicate entries,
# i.e., the matrix' attributes ``row``, ``col``, and ``data`` can contain
# entries that refer to the same matrix element, in which case they should be
# summed up. This is performed, e.g., during a conversion to CSR or CSC.
#
if hynet_sparse_(((1, 2), ((0, 0), (0, 0))), shape=(1, 1)).tocsr()[0, 0] != 3:
    raise RuntimeError("Sparse matrix initialization is incompatible.")

hynet_eps = np.finfo(hynet_float_).eps  # Machine epsilon


class BusType(Enum):
    """Type of voltage waveform experienced at the bus."""
    AC = 'ac'
    DC = 'dc'

    def __str__(self):
        return self.value.upper()


class BranchType(Enum):
    """Type of entity modeled by the branch."""
    LINE = 'line'
    TRANSFORMER = 'transformer'

    def __str__(self):
        return self.value.upper()


class InjectorType(Enum):
    """Type of entity modeled by the injector."""
    # Categories
    CONVENTIONAL = 'conventional'
    RENEWABLE = 'renewable'
    PROSUMER = 'prosumer'
    LOAD = 'load'
    COMPENSATION = 'compensation'

    # Subcategories for conventional generation
    COAL = 'conventional:coal'
    GAS = 'conventional:gas'
    NUCLEAR = 'conventional:nuclear'

    # Subcategories for renewables
    HYDRO = 'renewable:hydro'
    WIND = 'renewable:wind'
    PV = 'renewable:pv'
    BIOMASS = 'renewable:biomass'
    GEOTHERMAL = 'renewable:geothermal'

    def is_conventional(self):
        """Return ``True`` if it is a conventional generation utility."""
        return self.value.startswith('conventional')

    def is_renewable(self):
        """Return ``True`` if it is a renewables-based generation utility."""
        return self.value.startswith('renewable')

    def is_prosumer(self):
        """Return ``True`` if it is a prosumer."""
        return self.value.startswith('prosumer')

    def is_load(self):
        """Return ``True`` if it is a load."""
        return self.value.startswith('load')

    def is_compensation(self):
        """Return ``True`` if it is a reactive power compensator."""
        return self.value.startswith('compensation')

    def __str__(self):
        """
        Return a formatted string for the injector type.

        This formatted string is generated from the value of the element, which
        a hierarchy of categories separated by a colon, i.e.,
        ::
          category:subcategory:subsubcategory

        During formatting, any underscores are replaced by blanks.
        """
        categories = self.value.replace('_', ' ').split(':')
        type_string = categories.pop(0)
        if categories:
            type_string += ' (' + '/'.join(categories) + ')'
        return type_string.upper()


class EntityType(Enum):
    """Type of entity that is deactivated in a scenario."""
    BUS = 'bus'
    BRANCH = 'branch'
    CONVERTER = 'converter'
    SHUNT = 'shunt'
    INJECTOR = 'injector'

    def __str__(self):
        return self.value.upper()

    def __lt__(self, other):
        # REMARK: This operator is required by SQLAlchemy as EntityType is
        # part of the primary key in DBScenarioInactivity
        return self.value < other.value


class SolverType(Enum):
    """Type of problem that can be solved with the solver."""
    QCQP = 'QCQP'
    SDR = 'SDR'
    SOCR = 'SOCR'


class SolverStatus(Enum):
    """Status returned by a solver after performing an optimization."""
    SOLVED = 0
    INACCURATE = 1
    UNBOUNDED = -1
    INFEASIBLE = -2
    FAILED = -4


class DBInfoKey(Enum):
    """Valid keys for the ``db_info`` table in a *hynet* grid database."""
    VERSION = "version"
    GRID_NAME = "grid_name"
    BASE_MVA = "base_mva"
    DESCRIPTION = "description"

    def __str__(self):
        return self.value



#pylint: disable=too-many-lines,too-many-public-methods
"""
Steady-state system model of *hynet*.
"""

import logging

import numpy as np
from scipy.sparse import kron, coo_matrix

from hynet.types_ import hynet_int_, hynet_float_, hynet_complex_, hynet_sparse_
from hynet.utilities.base import (create_sparse_matrix,
                                  create_sparse_diag_matrix,
                                  create_sparse_zero_matrix,
                                  create_dense_vector,
                                  Timer)
from hynet.utilities.worker import workers
from hynet.utilities.graph import eliminate_parallel_edges
from hynet.qcqp.problem import QCQPPoint, ObjectiveFunction, Constraint, QCQP
from hynet.opf.result import OPFResult
from hynet.scenario.capability import ConverterCapRegion

_log = logging.getLogger(__name__)


def _call_and_concatenate(constraint_generator):
    """Calls the constraint generator and concatenates its result."""
    constraints = constraint_generator()
    if isinstance(constraints, tuple):
        constraints = np.concatenate(constraints)
    return constraints


class SystemModel:
    """
    System model for a steady-state scenario of a grid.

    Based on the specification of a scenario via a ``Scenario`` object, this
    class provides the methods to generate the corresponding system model
    equations and constraints as well as the associated optimal power flow
    (OPF) problem formulation. The state variables in this system model
    are the bus voltage vector ``v``, the converter state vector ``f``, the
    injector state vector ``s``, and the injector cost vector ``z``. The latter
    is actually not part of the system model but used for the reformulation of
    the piecewise linear active and reactive power cost functions of injectors.
    All state variables are considered in p.u., i.e., normalized.

    **Custom OPF Formulations:** The system model serves as a builder for the
    OPF problem. In order to customize the OPF formulation, one can define a
    corresponding system model by deriving from this class, implementing the
    additional constraint generators, registering them by overriding
    ``_get_constraint_generators``, and, if desired, overriding
    ``_get_opf_objective`` to customize the objective function. If additional
    variables are required, the auxiliary variable vector ``z`` may be expanded
    accordingly (override and increase ``dim_z`` and utilize the variables
    following the epigraph variables). If the customized OPF formulation
    necessitates a customized OPF result evaluation, implement the latter by
    deriving from ``OPFResult`` and associate this OPF result class with the
    system model by overriding ``create_opf_result``. To calculate a solution
    for the customized OPF, create an object of the specialized system model
    for the desired scenario and pass it to ``calc_opf``.

    See Also
    --------
    hynet.scenario.representation.Scenario:
        Specification of a steady-state grid scenario.
    hynet.opf.calc.calc_opf:
        Calculate an optimal power flow.
    """
    def __init__(self, scenario, verify_scenario=True):
        """
        Initialize the system model with the scenario data.

        Parameters
        ----------
        scenario : Scenario
            Steady-state scenario data for the model.
        verify_scenario : bool, optional
            If ``True`` (default), an integrity and validity check is performed
            on the provided scenario data (see ``Scenario.verify``). In case it
            is ensured beforehand that the provided data is consistent and
            valid, this check may be skipped to improve performance.
        """
        timer = Timer()
        if verify_scenario:
            scenario.verify()
        self._scr = scenario
        self._is_acyclic = self._scr.has_acyclic_subgrids()
        self._ac_subgrids = self._scr.get_ac_subgrids()
        self._dc_subgrids = self._scr.get_dc_subgrids()
        self._islands = self._scr.get_islands()
        (self._Y, self._Y_src, self._Y_dst) = \
            SystemModel._get_admittance_matrices(self._scr)
        self._cost_scaling = self._select_cost_scaling(scenario)
        _log.debug("Steady-state model creation ({:.3f} sec.)"
                   .format(timer.total()))

    @property
    def cost_function_scaling(self):
        """
        Return the cost function scaling used by ``get_cost_epigraph_constraints``.

        To improve the numerical conditioning of the OPF problem and,
        therewith, mitigate numerical issues with the solver, the cost
        functions are scaled. This scaling must be considered when including
        a loss term in the objective.
        """
        return self._cost_scaling

    @property
    def is_acyclic(self):
        """
        Return ``True`` if all subgrids are acyclic and ``False`` otherwise.

        **Remark:** This is determined at object construction and stored as a
        property for performance reasons.
        """
        return self._is_acyclic

    @property
    def ac_subgrids(self):
        """
        Return a list with a pandas index of bus IDs for every AC subgrid.

        **Remark:** This is determined at object construction and stored as a
        property for performance reasons.
        """
        return self._ac_subgrids

    @property
    def dc_subgrids(self):
        """
        Return a list with a pandas index of bus IDs for every DC subgrid.

        **Remark:** This is determined at object construction and stored as a
        property for performance reasons.
        """
        return self._dc_subgrids

    @property
    def islands(self):
        """
        Return a list with a pandas index of bus IDs for every islanded grid.

        **Remark:** This is determined at object construction and stored as a
        property for performance reasons.
        """
        return self._islands

    @staticmethod
    def _get_admittance_matrices(scenario):  # pylint: disable=too-many-locals
        """Return the bus, source, and destination admittance matrix."""
        N_E = scenario.num_branches
        N_V = scenario.num_buses

        e_src = scenario.e_src.values
        e_dst = scenario.e_dst.values

        rho_src = scenario.branch['rho_src'].values
        rho_dst = scenario.branch['rho_dst'].values
        rho = np.multiply(rho_src.conj(), rho_dst)

        y_tld = scenario.bus['y_tld'].values

        y_src = scenario.branch['y_src'].values
        y_dst = scenario.branch['y_dst'].values
        y_bar = 1 / scenario.branch['z_bar'].values

        alpha_src = np.multiply(np.square(np.abs(rho_src)), y_bar + y_src)
        alpha_dst = np.multiply(np.square(np.abs(rho_dst)), y_bar + y_dst)

        beta_src = -np.multiply(rho, y_bar)
        beta_dst = -np.multiply(rho.conj(), y_bar)

        alpha = y_tld + create_dense_vector(e_src, alpha_src, N_V) \
                      + create_dense_vector(e_dst, alpha_dst, N_V)

        Y_src = create_sparse_matrix(range(N_E), e_src, alpha_src, N_E, N_V) \
              + create_sparse_matrix(range(N_E), e_dst, beta_src, N_E, N_V)

        Y_dst = create_sparse_matrix(range(N_E), e_dst, alpha_dst, N_E, N_V) \
              + create_sparse_matrix(range(N_E), e_src, beta_dst, N_E, N_V)

        Y = create_sparse_diag_matrix(alpha) \
          + create_sparse_matrix(e_src, e_dst, beta_src, N_V, N_V) \
          + create_sparse_matrix(e_dst, e_src, beta_dst, N_V, N_V)

        # REMARK: The admittance matrices are primarily employed in row slicing
        # operations. For performance reasons, we thus use the CSR format here.
        return Y.tocsr(), Y_src.tocsr(), Y_dst.tocsr()

    @property
    def scenario(self):
        """Return the scenario data of the system model."""
        return self._scr

    @property
    def dim_v(self):
        """Return the dimension of the state variable ``v``."""
        return self._scr.num_buses

    @property
    def dim_f(self):
        """Return the dimension of the state variable ``f``."""
        return 4*self._scr.num_converters

    @property
    def dim_s(self):
        """Return the dimension of the state variable ``s``."""
        return 2*self._scr.num_injectors

    @property
    def dim_z(self):
        """Return the dimension of the state variable ``z``."""
        return 2*self._scr.num_injectors

    @property
    def Y(self):
        """Return the bus admittance matrix."""
        return self._Y

    @property
    def Y_src(self):
        """Return the source admittance matrix."""
        return self._Y_src

    @property
    def Y_dst(self):
        """Return the destination admittance matrix."""
        return self._Y_dst

    def get_normalization_factors(self):
        """
        Return the normalization factors of the state variables.

        Returns
        -------
        factors : QCQPPoint
            Contains the normalization factors of the state variables in the
            attributes ``v``, ``f``, ``s``, and ``z``.
        """
        return QCQPPoint(v=1.0,
                         f=1 / self._scr.base_mva,
                         s=1 / self._scr.base_mva,
                         z=self._cost_scaling)

    def _get_constraint_generators(self):
        """
        Return a list with all constraint generation functions.

        These constraint generation functions are used to generate the
        constraints of the OPF QCQP. Override this method in a derived class
        to **append** additional inequality constraints (The *first*
        constraint generator is assumed to return *equality constraints*,
        i.e., the power balance constraints, while all others are assumed to
        return *inequality constraints*, cf. ``get_opf_problem``).
        """
        return [self.get_balance_constraints,
                self.get_ampacity_constraints,
                self.get_real_part_constraints,
                self.get_angle_constraints,
                self.get_voltage_constraints,
                self.get_drop_constraints,
                self.get_converter_polyhedron_constraints,
                self.get_injector_polyhedron_constraints,
                self.get_cost_epigraph_constraints]

    def _get_opf_objective(self):
        """
        Return the objective function for the OPF problem formulation.

        Override this method in a derived class to customize the OPF objective.
        """
        N_I = self._scr.num_injectors

        if self._scr.loss_price != 0:
            (obj_C, obj_c) = self.get_dyn_loss_function()
            obj_C *= self._scr.loss_price * self._cost_scaling
            obj_c *= self._scr.loss_price * self._cost_scaling
        else:
            obj_C = create_sparse_zero_matrix(self.dim_v, self.dim_v)
            obj_c = create_sparse_zero_matrix(self.dim_f, 1)

        obj_r = np.zeros((self.dim_z, 1), dtype=hynet_float_)
        obj_r[:2*N_I, 0] = 1  # Cost function epigraph ordinate variables

        return ObjectiveFunction(C=obj_C,
                                 c=obj_c,
                                 a=create_sparse_zero_matrix(self.dim_s, 1),
                                 r=hynet_sparse_(obj_r),
                                 scaling=self._cost_scaling)

    def get_opf_problem(self):
        """
        Return the QCQP for the OPF problem of this system model.

        Returns
        -------
        qcqp : QCQP
            QCQP specification for the OPF problem associated with this
            system model.
        """
        timer = Timer()

        # Constraint generation (using parallel processing)
        constraints = workers.map(_call_and_concatenate,
                                  self._get_constraint_generators())

        # Variable bounds
        v_lb, v_ub = self.get_voltage_magnitude_bounds()
        f_lb, f_ub = self.get_converter_state_bounds()
        s_lb, s_ub = self.get_injector_state_bounds()
        z_lb, z_ub = self.get_cost_aux_var_bounds()

        qcqp = QCQP(obj_func=self._get_opf_objective(),
                    eq_crt=constraints[0],
                    ineq_crt=np.concatenate(constraints[1:]),
                    lb=QCQPPoint(v_lb, f_lb, s_lb, z_lb),
                    ub=QCQPPoint(v_ub, f_ub, s_ub, z_ub),
                    edges=eliminate_parallel_edges((self._scr.e_src.values,
                                                    self._scr.e_dst.values)),
                    roots=self._scr.get_ref_buses(),
                    normalization=self.get_normalization_factors())

        _log.debug("OPF QCQP creation ({:.3f} sec.)".format(timer.total()))
        return qcqp

    def create_opf_result(self, qcqp_result, total_time=np.nan):
        """
        Create and return an OPF result object.

        This method serves as a factory for an OPF result object using the OPF
        result class associated with this system model.

        Override this method in a derived class if the customized OPF
        formulation necessitates a customized result evaluation. The customized
        OPF class must be a specialization of ``OPFResult``.

        Parameters
        ----------
        qcqp_result : hynet.qcqp.result.QCQPResult
            Solution of the OPF QCQP.
        total_time : .hynet_float_, optional
            Total time for solving the OPF, cf. hynet.opf.calc.calc_opf.

        Returns
        -------
        result : hynet.opf.result.OPFResult
        """
        return OPFResult(self, qcqp_result, total_time=total_time)

    def get_balance_constraints(self):  # pylint: disable=too-many-locals
        """Return the active and reactive power balance constraints."""
        timer = Timer()
        N_V = self._scr.num_buses
        N_C = self._scr.num_converters
        N_I = self._scr.num_injectors
        index = self._scr.bus.index
        c_src = self._scr.c_src.values
        c_dst = self._scr.c_dst.values
        n_src = self._scr.n_src.values

        # CAVEAT: For better performance, the type conversion in the loop below
        # is made via .tocoo(). Update the code if the format has changed.
        assert hynet_sparse_ is coo_matrix

        load = self._scr.bus['load'].values / self._scr.base_mva

        loss_fix = self._scr.converter['loss_fix'].values / self._scr.base_mva
        loss_fwd = self._scr.converter['loss_fwd'].values / 100
        loss_bwd = self._scr.converter['loss_bwd'].values / 100

        # Accumulate static losses of converters at their respective source bus
        load += create_dense_vector(c_src, loss_fix, N_V)

        # Standard basis vectors in R^2 and R^4
        e1_R2 = create_sparse_matrix([0], [0], [1], 2, 1, dtype=hynet_float_)
        e2_R2 = create_sparse_matrix([1], [0], [1], 2, 1, dtype=hynet_float_)
        e1_R4 = create_sparse_matrix([0], [0], [1], 4, 1, dtype=hynet_float_)
        e2_R4 = create_sparse_matrix([1], [0], [1], 4, 1, dtype=hynet_float_)
        e3_R4 = create_sparse_matrix([2], [0], [1], 4, 1, dtype=hynet_float_)
        e4_R4 = create_sparse_matrix([3], [0], [1], 4, 1, dtype=hynet_float_)

        # Prepare calculation of vectors p_n and q_n
        A_src = create_sparse_matrix(range(N_C), c_src, np.ones(N_C), N_C, N_V,
                                     dtype=hynet_float_)
        A_dst = create_sparse_matrix(range(N_C), c_dst, np.ones(N_C), N_C, N_V,
                                     dtype=hynet_float_)
        B_src = create_sparse_matrix(range(N_C), c_src, 1 - loss_bwd, N_C, N_V,
                                     dtype=hynet_float_)
        B_dst = create_sparse_matrix(range(N_C), c_dst, 1 - loss_fwd, N_C, N_V,
                                     dtype=hynet_float_)

        # REMARK: CSC format for efficient column slicing
        c_p = (kron(A_src, e1_R4) - kron(B_src, e2_R4) +
               kron(A_dst, e2_R4) - kron(B_dst, e1_R4)).tocsc()
        c_q = (-kron(A_src, e3_R4) - kron(A_dst, e4_R4)).tocsc()

        # Prepare calculation of vectors for injectors
        H = create_sparse_matrix(range(N_I), n_src, -np.ones(N_I), N_I, N_V,
                                 dtype=hynet_float_)

        # REMARK: CSC format for efficient column slicing
        a_p = kron(H, e1_R2).tocsc()
        a_q = kron(H, e2_R2).tocsc()

        crt_p = np.ndarray(N_V, dtype=Constraint)
        crt_q = np.ndarray(N_V, dtype=Constraint)

        for n in range(N_V):
            # REMARK to the construction of the matrices P_n and Q_n: To keep
            # the number of temporary matrices minimal, the construction of
            # P_n = (S_n + S_n^H) / 2 and Q_n = (S_n - S_n^H) / 2j utilizes the
            # fact that S_n consists of the conjugate transpose of the n-th row
            # of Y in its n-th column.
            y_n = self._Y[n, :].tocoo()
            idx_row = np.concatenate([y_n.col, y_n.row + n])
            idx_col = np.concatenate([y_n.row + n, y_n.col])
            P_n = create_sparse_matrix(idx_row, idx_col,
                                       np.concatenate([y_n.data.conj(),
                                                       y_n.data]) / 2,
                                       N_V, N_V, dtype=hynet_complex_)
            Q_n = create_sparse_matrix(idx_row, idx_col,
                                       np.concatenate([y_n.data.conj(),
                                                       -y_n.data]) / 2j,
                                       N_V, N_V, dtype=hynet_complex_)

            # Active power balance
            crt_p[n] = Constraint(name='bal_p',
                                  table='bus',
                                  id=index[n],
                                  C=P_n,
                                  c=c_p[:, n].tocoo(),
                                  a=a_p[:, n].tocoo(),
                                  r=None,
                                  b=-load[n].real,
                                  scaling=1 / self._scr.base_mva)

            # Reactive power balance
            crt_q[n] = Constraint(name='bal_q',
                                  table='bus',
                                  id=index[n],
                                  C=Q_n,
                                  c=c_q[:, n].tocoo(),
                                  a=a_q[:, n].tocoo(),
                                  r=None,
                                  b=-load[n].imag,
                                  scaling=1 / self._scr.base_mva)

        _log.debug("Power balance constraints ({:.3f} sec.)"
                   .format(timer.total()))
        return crt_p, crt_q

    def get_voltage_constraints(self):
        """Return the voltage magnitude lower and upper bound constraints."""
        timer = Timer()
        N_V = self._scr.num_buses
        index = self._scr.bus.index

        v_min_squared = np.square(self._scr.bus['v_min'].values)
        v_max_squared = np.square(self._scr.bus['v_max'].values)

        crt_v_min = np.ndarray(N_V, dtype=Constraint)
        crt_v_max = np.ndarray(N_V, dtype=Constraint)

        for n in range(N_V):
            M_n = create_sparse_matrix([n], [n], [1], N_V, N_V,
                                       dtype=hynet_float_)

            # Lower bound
            crt_v_min[n] = Constraint(name='v_min',
                                      table='bus',
                                      id=index[n],
                                      C=-M_n,
                                      c=None,
                                      a=None,
                                      r=None,
                                      b=-v_min_squared[n],
                                      scaling=1.0)

            # Upper bound
            crt_v_max[n] = Constraint(name='v_max',
                                      table='bus',
                                      id=index[n],
                                      C=M_n,
                                      c=None,
                                      a=None,
                                      r=None,
                                      b=v_max_squared[n],
                                      scaling=1.0)

        _log.debug("Voltage constraints ({:.3f} sec.)"
                   .format(timer.total()))
        return crt_v_min, crt_v_max

    def get_ampacity_constraints(self):
        """Return the source and destination ampacity constraints."""
        timer = Timer()
        N_E = self._scr.num_branches
        index = self._scr.branch.index

        # CAVEAT: For better performance, the type conversion in the loop below
        # is made via .tocoo(). Update the code if the format has changed.
        assert hynet_sparse_ is coo_matrix

        # Maximum current is defined via apparent power flow rating at 1 p.u.
        rating = self._scr.branch['rating'].values
        I_max_squared = np.square(rating / self._scr.base_mva)

        # CONSTRAINT SCALING: In the description of the hybrid AC/DC power
        # grid model, it can be observed that the elements of all constraint
        # matrices are either roughly close to 1 or in the range of the series
        # admittance of the AC branches, except for the current constraints,
        # which feature the series admittance squared. As the series impedance
        # may be very small, the corresponding series admittance is large,
        # e.g., in the range of 10^4. Then, the modulus of the constraint
        # matrix elements is in the range between 0 and 10^4, except for the
        # current constraints, which then feature elements in the range of
        # 10^8. We observed that this frequently leads to numerical issues in
        # the employed solver. To reduce the dynamic range observed in the
        # constraint matrices and, therewith, improve the conditioning of the
        # problem, the current constraints are scaled by the minimum branch
        # series impedance.
        if N_E != 0:
            scaling = np.min(np.abs(self._scr.branch['z_bar'].values))

        # Allocate arrays for (non-omitted) constraint objects
        M = np.count_nonzero(~np.isnan(I_max_squared))
        crt_i_max_src = np.ndarray(M, dtype=Constraint)
        crt_i_max_dst = np.ndarray(M, dtype=Constraint)

        m = 0
        for k in range(N_E):
            if np.isnan(I_max_squared[k]):
                continue

            # Flow limit at source bus
            I_src_k = self._Y_src[k, :].conj().transpose().dot(self._Y_src[k, :])
            crt_i_max_src[m] = Constraint(name='i_max_src',
                                          table='branch',
                                          id=index[k],
                                          C=(I_src_k * scaling).tocoo(),
                                          c=None,
                                          a=None,
                                          r=None,
                                          b=I_max_squared[k] * scaling,
                                          scaling=scaling)

            # Flow limit at destination bus
            I_dst_k = self._Y_dst[k, :].conj().transpose().dot(self._Y_dst[k, :])
            crt_i_max_dst[m] = Constraint(name='i_max_dst',
                                          table='branch',
                                          id=index[k],
                                          C=(I_dst_k * scaling).tocoo(),
                                          c=None,
                                          a=None,
                                          r=None,
                                          b=I_max_squared[k] * scaling,
                                          scaling=scaling)
            m += 1

        _log.debug("Ampacity constraints ({:.3f} sec.)"
                   .format(timer.total()))
        return crt_i_max_src, crt_i_max_dst

    def get_drop_constraints(self):
        """Return the voltage drop lower and upper bound constraints."""
        timer = Timer()
        N_V = self._scr.num_buses
        N_E = self._scr.num_branches
        index = self._scr.branch.index
        e_src = self._scr.e_src.values
        e_dst = self._scr.e_dst.values

        nu_min = self._scr.branch['drop_min'].values / 100
        nu_max = self._scr.branch['drop_max'].values / 100

        crt_drop_min = np.ndarray(np.count_nonzero(~np.isnan(nu_min)),
                                  dtype=Constraint)
        crt_drop_max = np.ndarray(np.count_nonzero(~np.isnan(nu_max)),
                                  dtype=Constraint)

        m_min = m_max = 0
        for k in range(N_E):
            src_dst = [e_src[k], e_dst[k]]

            # Lower bound on voltage drop
            if not np.isnan(nu_min[k]):
                M_min_k = create_sparse_matrix(src_dst, src_dst,
                                               [(1+nu_min[k]) ** 2, -1],
                                               N_V, N_V, dtype=hynet_float_)
                crt_drop_min[m_min] = Constraint(name='drop_min',
                                                 table='branch',
                                                 id=index[k],
                                                 C=M_min_k,
                                                 c=None,
                                                 a=None,
                                                 r=None,
                                                 b=0.0,
                                                 scaling=1.0)
                m_min += 1

            # Upper bound on voltage drop
            if not np.isnan(nu_max[k]):
                M_max_k = create_sparse_matrix(src_dst, src_dst,
                                               [-(1+nu_max[k]) ** 2, 1],
                                               N_V, N_V, dtype=hynet_float_)
                crt_drop_max[m_max] = Constraint(name='drop_max',
                                                 table='branch',
                                                 id=index[k],
                                                 C=M_max_k,
                                                 c=None,
                                                 a=None,
                                                 r=None,
                                                 b=0.0,
                                                 scaling=1.0)
                m_max += 1

        _log.debug("Voltage drop constraints ({:.3f} sec.)"
                   .format(timer.total()))
        return crt_drop_min, crt_drop_max

    def get_real_part_constraints(self):
        """
        Return the "real-part" constraints.

        These constraints ensure a precondition assumed by the voltage angle
        difference constraints, i.e., that the voltage angle difference is
        limited to +/- 90 degrees.
        """
        timer = Timer()
        N_V = self._scr.num_buses
        N_E = self._scr.num_branches
        index = self._scr.branch.index
        e_src = self._scr.e_src.values
        e_dst = self._scr.e_dst.values

        crt_real_part = np.ndarray(N_E, dtype=Constraint)

        for k in range(N_E):
            A_k = create_sparse_matrix([e_src[k], e_dst[k]],
                                       [e_dst[k], e_src[k]],
                                       [-1, -1], N_V, N_V, dtype=hynet_float_)
            crt_real_part[k] = Constraint(name='real_part',
                                          table='branch',
                                          id=index[k],
                                          C=A_k,
                                          c=None,
                                          a=None,
                                          r=None,
                                          b=0.0,
                                          scaling=1.0)

        _log.debug("Angle 'real-part' constraints ({:.3f} sec.)"
                   .format(timer.total()))
        return crt_real_part

    def get_angle_constraints(self):
        """
        Return the angle diff. lower and upper bound and "real-part" constraints.
        """
        timer = Timer()
        N_V = self._scr.num_buses
        N_E = self._scr.num_branches
        index = self._scr.branch.index
        e_src = self._scr.e_src.values
        e_dst = self._scr.e_dst.values

        tan_delta_min = np.tan(self._scr.branch['angle_min'].values * np.pi/180)
        tan_delta_max = np.tan(self._scr.branch['angle_max'].values * np.pi/180)

        crt_angle_min = np.ndarray(np.count_nonzero(~np.isnan(tan_delta_min)),
                                   dtype=Constraint)
        crt_angle_max = np.ndarray(np.count_nonzero(~np.isnan(tan_delta_max)),
                                   dtype=Constraint)

        m_min = m_max = 0
        for k in range(N_E):
            src_dst = [e_src[k], e_dst[k]]
            dst_src = src_dst[::-1]

            # Lower bound on angle difference
            if not np.isnan(tan_delta_min[k]):
                A_min_k = create_sparse_matrix(src_dst, dst_src,
                                               [tan_delta_min[k]+1j,
                                                tan_delta_min[k]-1j],
                                               N_V, N_V, dtype=hynet_complex_)
                crt_angle_min[m_min] = Constraint(name='angle_min',
                                                  table='branch',
                                                  id=index[k],
                                                  C=A_min_k,
                                                  c=None,
                                                  a=None,
                                                  r=None,
                                                  b=0.0,
                                                  scaling=1.0)
                m_min += 1

            # Upper bound on angle difference
            if not np.isnan(tan_delta_max[k]):
                A_max_k = create_sparse_matrix(src_dst, dst_src,
                                               [-tan_delta_max[k]-1j,
                                                -tan_delta_max[k]+1j],
                                               N_V, N_V, dtype=hynet_complex_)
                crt_angle_max[m_max] = Constraint(name='angle_max',
                                                  table='branch',
                                                  id=index[k],
                                                  C=A_max_k,
                                                  c=None,
                                                  a=None,
                                                  r=None,
                                                  b=0.0,
                                                  scaling=1.0)
                m_max += 1

        _log.debug("Angle difference constraints ({:.3f} sec.)"
                   .format(timer.total()))
        return crt_angle_min, crt_angle_max

    def get_converter_polyhedron_constraints(self):
        """Return the converter capability region polyhedron constraints."""
        timer = Timer()
        N_C = self._scr.num_converters
        index = self._scr.converter.index
        cap_src = self._scr.converter['cap_src'].values
        cap_dst = self._scr.converter['cap_dst'].values
        loss_fwd = self._scr.converter['loss_fwd'].values / 100
        loss_bwd = self._scr.converter['loss_bwd'].values / 100

        crt_conv_poly = np.ndarray((0,), dtype=Constraint)
        idx_col = np.zeros(4, dtype=hynet_int_)

        for l in range(N_C):
            A_src, b_src, name_src = cap_src[l].get_polyhedron('src', loss_bwd[l])
            A_dst, b_dst, name_dst = cap_dst[l].get_polyhedron('dst', loss_fwd[l])
            A = np.vstack((A_src, A_dst))
            b = np.concatenate((b_src, b_dst))
            name = ['src_' + x for x in name_src] + ['dst_' + x for x in name_dst]

            crt = np.ndarray(len(b), dtype=Constraint)
            idx_row = np.arange(4*l, 4*(l+1))

            # Create constraint objects for A*f_l <= b
            # (As f_l is normalized by base_mva, b is adjusted accordingly)
            for i in range(len(b)):
                crt[i] = Constraint(name='cap_' + name[i],
                                    table='converter',
                                    id=index[l],
                                    C=None,
                                    c=create_sparse_matrix(idx_row, idx_col,
                                                           A[i, :], 4*N_C, 1,
                                                           dtype=hynet_float_),
                                    a=None,
                                    r=None,
                                    b=b[i] / self._scr.base_mva,
                                    scaling=1 / self._scr.base_mva)

            crt_conv_poly = np.concatenate((crt_conv_poly, crt))

        _log.debug("Converter polyhedron constraints ({:.3f} sec.)"
                   .format(timer.total()))
        return crt_conv_poly

    def get_injector_polyhedron_constraints(self):
        """Return the injector capability region polyhedron constraints."""
        timer = Timer()
        N_I = self._scr.num_injectors
        index = self._scr.injector.index
        cap = self._scr.injector['cap'].values

        crt_inj_poly = np.ndarray((0,), dtype=Constraint)

        for j in range(N_I):
            A, b, name = cap[j].get_polyhedron()
            crt = np.ndarray(len(b), dtype=Constraint)

            # Create constraint objects for A*s_j <= b
            # (As s_j is normalized by base_mva, b is adjusted accordingly)
            for i in range(len(b)):
                crt[i] = Constraint(name='cap_' + name[i],
                                    table='injector',
                                    id=index[j],
                                    C=None,
                                    c=None,
                                    a=create_sparse_matrix([2*j, 2*j + 1],
                                                           [0, 0], A[i, :],
                                                           2*N_I, 1,
                                                           dtype=hynet_float_),
                                    r=None,
                                    b=b[i] / self._scr.base_mva,
                                    scaling=1 / self._scr.base_mva)

            crt_inj_poly = np.concatenate((crt_inj_poly, crt))

        _log.debug("Injector polyhedron constraints ({:.3f} sec.)"
                   .format(timer.total()))
        return crt_inj_poly

    def get_cost_epigraph_constraints(self):
        """Return the PWL cost function epigraph constraints."""
        timer = Timer()
        N_I = self._scr.num_injectors
        index = self._scr.injector.index
        cost_p = self._scr.injector['cost_p'].values
        cost_q = self._scr.injector['cost_q'].values

        crt_inj_epi = np.ndarray((0,), dtype=Constraint)
        a_scaling = self._scr.base_mva * self._cost_scaling

        for j in range(N_I):
            if cost_p[j] is not None:
                (A, b) = cost_p[j].get_epigraph_polyhedron()
                crt = np.ndarray(len(b), dtype=Constraint)

                # Create constraint objects for z_p_l*1 >= A*s_p_l + b
                for i in range(len(b)):
                    crt[i] = Constraint(name='inj_p_epi_' + str(i),
                                        table=None,
                                        id=index[j],
                                        C=None,
                                        c=None,
                                        a=create_sparse_matrix([2*j], [0],
                                                               [A[i] * a_scaling],
                                                               self.dim_s, 1,
                                                               dtype=hynet_float_),
                                        r=create_sparse_matrix([2*j], [0], [-1],
                                                               self.dim_z, 1,
                                                               dtype=hynet_float_),
                                        b=-b[i] * self._cost_scaling,
                                        scaling=self._cost_scaling)

                crt_inj_epi = np.concatenate((crt_inj_epi, crt))

            if cost_q[j] is not None:
                (A, b) = cost_q[j].get_epigraph_polyhedron()
                crt = np.ndarray(len(b), dtype=Constraint)

                # Create constraint objects for z_q_l*1 >= A*s_q_l + b
                for i in range(len(b)):
                    crt[i] = Constraint(name='inj_q_epi_' + str(i),
                                        table=None,
                                        id=index[j],
                                        C=None,
                                        c=None,
                                        a=create_sparse_matrix([2*j + 1], [0],
                                                               [A[i] * a_scaling],
                                                               self.dim_s, 1,
                                                               dtype=hynet_float_),
                                        r=create_sparse_matrix([2*j + 1], [0], [-1],
                                                               self.dim_z, 1,
                                                               dtype=hynet_float_),
                                        b=-b[i] * self._cost_scaling,
                                        scaling=self._cost_scaling)

                crt_inj_epi = np.concatenate((crt_inj_epi, crt))

        _log.debug("Cost function epigraph constraints ({:.3f} sec.)"
                   .format(timer.total()))
        return crt_inj_epi

    def get_voltage_magnitude_bounds(self):
        """
        Return the voltage magnitude bounds ``v_lb <= |v| <= v_ub``.

        **Remark:** The voltage magnitude bounds are captured by
        ``get_voltage_constraints`` and, thus, this box constraint is actually
        redundant and only included to provide optimization variable bounds
        for the solver (which, for some solvers, can improve convergence).
        To avoid any impact on the dual variables of the voltage magnitude
        constraints (which e.g. was observed with MOSEK), these box constraints
        are loosened w.r.t. the limits employed in ``get_voltage_constraints``.
        """
        gap = 0.1 * self._scr.bus['v_max'].values  # Loosened by 10% of the UB
        v_lb = self._scr.bus['v_min'].values - gap
        v_ub = self._scr.bus['v_max'].values + gap
        return v_lb, v_ub

    def get_converter_state_bounds(self):
        """Return the converter state bounds ``f_lb <= f <= f_ub``."""
        N_C = self._scr.num_converters

        cap_src = self._scr.converter['cap_src'].values
        cap_dst = self._scr.converter['cap_dst'].values

        f_lb = np.zeros(4*N_C, dtype=hynet_float_)
        f_ub = np.zeros(4*N_C, dtype=hynet_float_)

        for l in range(N_C):
            (f_lb[4*l:4*(l+1)], f_ub[4*l:4*(l+1)]) = \
                ConverterCapRegion.get_box_constraint(cap_src[l], cap_dst[l])

        return f_lb / self._scr.base_mva, f_ub / self._scr.base_mva

    def get_injector_state_bounds(self):
        """Return the injector state bounds ``s_lb <= s <= s_ub``."""
        N_I = self._scr.num_injectors

        cap = self._scr.injector['cap'].values

        s_lb = np.zeros(2*N_I, dtype=hynet_float_)
        s_ub = np.zeros(2*N_I, dtype=hynet_float_)

        for j in range(N_I):
            s_lb[2*j:2*(j+1)] = (cap[j].p_min, cap[j].q_min)
            s_ub[2*j:2*(j+1)] = (cap[j].p_max, cap[j].q_max)

        return s_lb / self._scr.base_mva, s_ub / self._scr.base_mva

    def get_cost_aux_var_bounds(self):
        """
        Return the auxiliary cost variable bounds ``z_lb <= z <= z_ub``.

        The lower and upper bound is set to ``numpy.nan`` if the corresponding
        bound should be omitted.
        """
        N_I = self._scr.num_injectors

        cost_p = self._scr.injector['cost_p'].values
        cost_q = self._scr.injector['cost_q'].values

        z_lb = np.nan * np.ones(self.dim_z, dtype=hynet_float_)
        z_ub = np.nan * np.ones(self.dim_z, dtype=hynet_float_)

        for j in range(N_I):
            if cost_p[j] is None:
                z_lb[2*j] = z_ub[2*j] = 0
            if cost_q[j] is None:
                z_lb[2*j + 1] = z_ub[2*j + 1] = 0

        return z_lb, z_ub

    def get_dyn_loss_function(self):
        """
        Return ``(C, c)`` for the total dyn. loss ``L(v,f) = v^H C v + c^T f``.
        """
        e1_R4 = create_sparse_matrix([0], [0], [1], 4, 1, dtype=hynet_float_)
        e2_R4 = create_sparse_matrix([1], [0], [1], 4, 1, dtype=hynet_float_)

        loss_fwd = self._scr.converter['loss_fwd'].values / 100
        loss_bwd = self._scr.converter['loss_bwd'].values / 100

        C = (self._Y + self._Y.conj().transpose()) / 2
        c = kron(loss_fwd.reshape((-1, 1)), e1_R4) + \
            kron(loss_bwd.reshape((-1, 1)), e2_R4)
        return (hynet_sparse_(C * self._scr.base_mva),
                hynet_sparse_(c * self._scr.base_mva))

    @staticmethod
    def _select_cost_scaling(scenario):
        """
        Return a suitable scaling factor for the objective of the OPF problem.
        """
        cost_p = scenario.injector['cost_p'].values
        cost_q = scenario.injector['cost_q'].values
        cap = scenario.injector['cap'].values
        load = scenario.bus['load'].values

        max_cost = 0
        for (f_p, f_q, cr) in zip(cost_p, cost_q, cap):
            if f_p is not None:
                max_cost = np.max([max_cost,
                                   np.max(np.abs(
                                       f_p.evaluate([cr.p_min, cr.p_max])))])
            if f_q is not None:
                max_cost = np.max([max_cost,
                                   np.max(np.abs(
                                       f_q.evaluate([cr.q_min, cr.q_max])))])

        if max_cost == 0:
            # Looks like we are facing loss minimization...
            if scenario.loss_price > 0:
                return 1 / (scenario.loss_price * scenario.base_mva)
            else:
                return 1.0

        max_inj = np.max([1,
                          np.max([np.abs([x.p_min, x.p_max,
                                          x.q_min, x.q_max]).max()
                                  for x in cap]) / scenario.base_mva,
                          np.max([np.abs([x.real, x.imag]).max()
                                  for x in load]) / scenario.base_mva
                          ])

        if max_cost > max_inj:
            return max_inj / max_cost
        return 1

    def calc_dynamic_losses(self, operating_point):
        """
        Return the total dynamic losses in MW for the given operating point.

        Parameters
        ----------
        operating_point : QCQPPoint
            Operating point of the system *without* the normalization of the
            state variables (except the voltage), i.e., as provided by an
            OPF result.
        """
        v = operating_point.v
        f = operating_point.f / self._scr.base_mva
        (C, c) = self.get_dyn_loss_function()
        return np.asscalar(v.conj().dot(C.dot(v)) + c.transpose().dot(f)).real

    def calc_shunt_apparent_power(self, operating_point):
        """
        Return the shunt apparent power in MVA for the given operating point.

        Parameters
        ----------
        operating_point : QCQPPoint
            Operating point of the system *without* the normalization of the
            state variables (except the voltage), i.e., as provided by an
            OPF result.
        """
        v = operating_point.v
        y_tld = self._scr.bus['y_tld'].values
        s_shunt = np.multiply(y_tld.conj(), np.square(np.abs(v)))
        return s_shunt * self._scr.base_mva

    def calc_branch_flow(self, operating_point):
        """
        Return the flow on the branches for the given operating point.

        Parameters
        ----------
        operating_point : QCQPPoint
            Operating point of the system *without* the normalization of the
            state variables (except the voltage), i.e., as provided by an
            OPF result.

        Returns
        -------
        i_src : numpy.ndarray[.hynet_float_]
            Branch current flow in p.u. at the source bus.
        i_dst : numpy.ndarray[.hynet_float_]
            Branch current flow in p.u. at the destination bus.
        s_src : numpy.ndarray[.hynet_complex_]
            Branch apparent power flow in MVA at the source bus.
        s_dst : numpy.ndarray[.hynet_complex_]
            Branch apparent power flow in MVA at the destination bus.
        """
        v = operating_point.v
        e_src = self._scr.e_src.values
        e_dst = self._scr.e_dst.values

        i_src = self._Y_src.dot(v)
        i_dst = self._Y_dst.dot(v)
        s_src = np.multiply(v[e_src], i_src.conj()) * self._scr.base_mva
        s_dst = np.multiply(v[e_dst], i_dst.conj()) * self._scr.base_mva
        return i_src, i_dst, s_src, s_dst

    def calc_branch_voltage_drop(self, operating_point):
        """
        Return the relative voltage magnitude drop along the branches.

        Parameters
        ----------
        operating_point : QCQPPoint
            Operating point of the system *without* the normalization of the
            state variables (except the voltage), i.e., as provided by an
            OPF result.
        """
        v = operating_point.v
        e_src = self._scr.e_src.values
        e_dst = self._scr.e_dst.values
        v_drop = np.true_divide(np.abs(v[e_dst]), np.abs(v[e_src])) - 1
        return v_drop

    def calc_branch_angle_difference(self, operating_point):
        """
        Return the voltage angle difference along the branches in degrees.

        Parameters
        ----------
        operating_point : QCQPPoint
            Operating point of the system *without* the normalization of the
            state variables (except the voltage), i.e., as provided by an
            OPF result.
        """
        v = operating_point.v
        e_src = self._scr.e_src.values
        e_dst = self._scr.e_dst.values
        angle_diff = np.angle(np.multiply(v[e_src].conj(), v[e_dst])) * 180/np.pi
        return angle_diff

    def calc_branch_effective_rating(self, operating_point):
        """
        Return the ampacity rating in MVA rating at the current bus voltages.

        In the scenario data, the branch flow is limited by an ampacity rating
        that is specified in terms of a long-term MVA rating at a bus voltage
        of 1 p.u.. Correspondingly, the ampacity rating translates to a
        different MVA rating if the bus voltages differ from 1 p.u.. This
        function returns the ampacity rating in MVA at the current bus
        voltages, i.e., the rating at 1 p.u. times the actual voltage.

        Parameters
        ----------
        operating_point : QCQPPoint
            Operating point of the system *without* the normalization of the
            state variables (except the voltage), i.e., as provided by an
            OPF result.
        """
        v_abs = np.abs(operating_point.v)
        e_src = self._scr.e_src.values
        e_dst = self._scr.e_dst.values
        effective_rating = np.multiply(self._scr.branch['rating'].values,
                                       np.maximum(v_abs[e_src], v_abs[e_dst]))
        return effective_rating

    def calc_converter_flow(self, operating_point):
        """
        Return the apparent power flow into the converters.

        Parameters
        ----------
        operating_point : QCQPPoint
            Operating point of the system *without* the normalization of the
            state variables (except the voltage), i.e., as provided by an
            OPF result.

        Returns
        -------
        s_src : numpy.ndarray[.hynet_complex_]
            Apparent power flow in MVA into the converter at the source bus.
        s_dst : numpy.ndarray[.hynet_complex_]
            Apparent power flow in MVA into the converter at the destination bus.
        """
        p_src = operating_point.f[0::4]
        p_dst = operating_point.f[1::4]
        q_src = operating_point.f[2::4]
        q_dst = operating_point.f[3::4]
        loss_fwd = self._scr.converter['loss_fwd'].values / 100
        loss_bwd = self._scr.converter['loss_bwd'].values / 100

        s_src = p_src - np.multiply(1 - loss_bwd, p_dst) - 1j*q_src
        s_dst = p_dst - np.multiply(1 - loss_fwd, p_src) - 1j*q_dst
        return s_src, s_dst

    def calc_injection_costs(self, operating_point):
        """
        Return the injector's active and reactive power injection costs in dollars.

        Parameters
        ----------
        operating_point : QCQPPoint
            Operating point of the system *without* the normalization of the
            state variables (except the voltage), i.e., as provided by an
            OPF result.

        Returns
        -------
        cost_p : numpy.ndarray[.hynet_complex_]
            Cost of the active power injection in dollars.
        cost_q : numpy.ndarray[.hynet_complex_]
            Cost of the reactive power injection in dollars.
        """
        s = operating_point.s[0::2] + 1j*operating_point.s[1::2]
        cost_p = np.nan * np.ones(self._scr.num_injectors, dtype=hynet_float_)
        cost_q = np.nan * np.ones(self._scr.num_injectors, dtype=hynet_float_)
        cost_iter = zip(s,
                        self._scr.injector['cost_p'].values,
                        self._scr.injector['cost_q'].values)

        for j, (s_j, f_p, f_q) in enumerate(cost_iter):
            if f_p is not None:
                cost_p[j] = f_p.evaluate(s_j.real)
            if f_q is not None:
                cost_q[j] = f_q.evaluate(s_j.imag)
        return cost_p, cost_q



#pylint: disable=anomalous-backslash-in-string
"""
Representation of a *hynet*-specific QCQP.
"""

import logging
from collections import namedtuple
import copy

import numpy as np
import scipy.sparse

from hynet.types_ import hynet_int_, hynet_float_, hynet_sparse_
from hynet.utilities.worker import workers
from hynet.utilities.base import (create_sparse_matrix,
                                  create_sparse_diag_matrix,
                                  Timer)

_log = logging.getLogger(__name__)


class QCQPPoint:
    """
    Point in the space of the optimization variables of the QCQP problem.

    Parameters
    ----------
    v : numpy.ndarray[.hynet_complex_] or numpy.ndarray[.hynet_float_]
    f : numpy.ndarray[.hynet_float_]
    s : numpy.ndarray[.hynet_float_]
    z : numpy.ndarray[.hynet_float_]
    """
    def __init__(self, v, f, s, z):
        self.v = v
        self.f = f
        self.s = s
        self.z = z

    def scale(self, scaling):
        """
        Scale the point by ``scaling``.

        Parameters
        ----------
        scaling : QCQPPoint or .hynet_float_
            Scaling factor, which can either be a numeric value, to scale all
            variables equally, or a QCQP point, to scale the variables
            individually, where the attributes ``v``, ``f``, ``s``, and ``z``
            specify the scaling factor for the corresponding variable.
        """
        if isinstance(scaling, QCQPPoint):
            scaling_v = scaling.v
            scaling_f = scaling.f
            scaling_s = scaling.s
            scaling_z = scaling.z
        else:
            scaling_v = scaling_f = scaling_s = scaling_z = scaling
        if self.v is not None:
            self.v *= scaling_v
        if self.f is not None:
            self.f *= scaling_f
        if self.s is not None:
            self.s *= scaling_s
        if self.z is not None:
            self.z *= scaling_z
        return self

    def reciprocal(self):
        """Set the point to the element-wise reciprocal of the variables."""
        if self.v is not None:
            self.v = 1 / self.v
        if self.f is not None:
            self.f = 1 / self.f
        if self.s is not None:
            self.s = 1 / self.s
        if self.z is not None:
            self.z = 1 / self.z
        return self

    def copy(self):
        """Return a deep copy of this point."""
        return copy.deepcopy(self)

    def __repr__(self):
        return ("QCQPPoint("
                + "v=" + repr(self.v) + ", "
                + "f=" + repr(self.f) + ", "
                + "s=" + repr(self.s) + ", "
                + "z=" + repr(self.z) + ")")


class ObjectiveFunction:
    """
    Specifies an objective ``g(v,f,s,z) = v^H C v + c^T f + a^T s + r^T z``.

    Parameters
    ----------
    C : .hynet_sparse_
        Hermitian matrix ``C`` of the objective function.
    c : .hynet_sparse_
        Vector ``c`` of the objective function.
    a : .hynet_sparse_
        Vector ``a`` of the objective function.
    r : .hynet_sparse_
        Vector ``r`` of the objective function.
    scaling : .hynet_float_
        Scaling factor for the objective.
    """
    def __init__(self, C, c, a, r, scaling):
        self.C = C
        self.c = c
        self.a = a
        self.r = r
        self.scaling = scaling

    def evaluate(self, p):
        """
        Evaluate the function, i.e., return ``y = g(v, f, s, z)``.

        Parameters
        ----------
        p : QCQPPoint
            Point that specifies ``v``, ``f``, ``s``, and ``z``.

        Returns
        -------
        y : .hynet_float_
            Function value ``y`` at the given point ``p``.
        """
        return np.asscalar(p.v.conj().transpose().dot(self.C.dot(p.v)).real
                           + self.c.transpose().dot(p.f)
                           + self.a.transpose().dot(p.s)
                           + self.r.transpose().dot(p.z))

    def __repr__(self):
        return ("ObjectiveFunction("
                + "C=" + ("None" if self.C is None else "<mtx>") + ", "
                + "c=" + ("None" if self.c is None else "<vec>") + ", "
                + "a=" + ("None" if self.a is None else "<vec>") + ", "
                + "r=" + ("None" if self.r is None else "<vec>")
                + " | scaling={:.3e})".format(self.scaling))


class Constraint(namedtuple('ConstraintBase', ['name', 'table', 'id',
                                               'C', 'c', 'a', 'r', 'b',
                                               'scaling'])):
    """
    Specifies a constraint ``v^H C v + c^T f + a^T s + r^T z <=/== b``.

    Parameters
    ----------
    name : str
        Name of the constraint and column name for the result table generation
        in ``QCQPResult``.
    table : str or None
        Name of the table with which this constraint is associated. This is
        used to generate the result tables in ``QCQPResult``. Set to ``None``
        to ignore the associated values during the result table generation.
    id : .hynet_id_
        Identifier of the row with which this constraint is associated. This is
        used to generate the result tables in ``QCQPResult``.
    C : .hynet_sparse_ or None
        Hermitian matrix ``C`` of the constraint function or ``None`` if this
        term is omitted.
    c : .hynet_sparse_ or None
        Vector ``c`` of the constraint function or ``None`` if this term is
        omitted.
    a : .hynet_sparse_ or None
        Vector ``a`` of the constraint function or ``None`` if this term is
        omitted.
    r : .hynet_sparse_ or None
        Vector ``r`` of the constraint function or ``None`` if this term is
        omitted.
    b : .hynet_float_
        Right-hand side scalar ``b`` of the constraint.
    scaling : .hynet_float_
        Scaling factor of the constraint.
    """
    def evaluate(self, p):
        """
        Return ``y = v^H C v + c^T f + a^T s + r^T z - b``.

        Parameters
        ----------
        p : QCQPPoint
            Point that specifies ``v``, ``f``, ``s``, and ``z``.

        Returns
        -------
        y : .hynet_float_
            Constraint value ``y`` at the given point ``p``.
        """
        result = -self.b
        if self.C is not None:
            result += np.asscalar(p.v.conj().transpose().dot(self.C.dot(p.v)).real)
        if self.c is not None:
            result += np.asscalar(self.c.transpose().dot(p.f))
        if self.a is not None:
            result += np.asscalar(self.a.transpose().dot(p.s))
        if self.r is not None:
            result += np.asscalar(self.r.transpose().dot(p.z))
        return result

    def __repr__(self):
        return "{0.name}:{0.id}".format(self)


class QCQP:
    """
    Specification of a quadratically constrained quadratic program::

            min   v^H C v + c^T f + a^T s + r^T z
          v,f,s,z

            s.t.  v^H C'_i v + c'_i^T f + a'_i^T s + r'_i^T z == b'_i, i = 1,...,N
                  v^H C"_j v + c"_j^T f + a"_j^T s + r"_j^T z <= b"_j, j = 1,...,M
                  v_lb <= |v| <= v_ub
                  f_lb <=  f  <= f_ub
                  s_lb <=  s  <= s_ub
                  z_lb <=  z  <= z_ub

    The vector ``v`` is complex-valued, while ``f``, ``s``, and ``z`` are
    real-valued. The bounds on ``|v|`` are optional to implement by the solver.
    They are assumed to be captured by the inequality constraints, while
    ``v_lb`` and ``v_ub`` may be used to support convergence of the solver.
    In contrary, the bounds on ``f``, ``s``, and ``z`` are mandatory to
    implement. A bound is ignored if the respective element is set to
    ``numpy.nan``.

    The objective function, constraint functions, and optimization variables
    may be scaled to improve the numerical conditioning of the problem. All
    coefficient matrices, coefficient vectors, bounds as well as the initial
    point, if provided, must be specified for the *scaled* problem. The scaling
    factors for the objective and the constraints can be specified via their
    ``scaling`` attribute, while the scaling of the optimization variables is
    defined via ``normalization`` (see below). This scaling information is
    utilized in ``QCQPResult`` to adjust the optimizer, the optimal value, and
    the dual variables to represent the solution of the unscaled problem.

    **Caution:** If the QCQP object is used in repeated computations with
    modifications of the constraints, the constraint vectorization buffering
    must be deactivated, see ``use_buffered_vectorization``.

    Parameters
    ----------
    obj_func : ObjectiveFunction
        Specification of the objective function.
    eq_crt : numpy.ndarray[Constraint]
        Specification of the equality constraints.
    ineq_crt : numpy.ndarray[Constraint]
        Specification of the inequality constraints.
    lb : QCQPPoint
        Specification of the lower bounds on ``|v|``, ``f``, ``s``, and ``z``.
    ub : QCQPPoint
        Specification of the upper bounds on ``|v|``, ``f``, ``s``, and ``z``.
    edges : (numpy.ndarray[.hynet_int_], numpy.ndarray[.hynet_int_])
        Specification of the sparsity pattern of the constraint matrices via
        the edges ``(edge_src, edge_dst)`` of an undirected graph. The constraint
        matrices may only have nonzero entries in row/column ``edge_src[i]`` and
        column/row ``edge_dst[i]``, with ``i in range(K)`` where
        ``K = len(edge_src) = len(edge_dst)``, as well as on the diagonal. The
        edges specified in ``edges`` must be *unique*, i.e., there must not exist
        any parallel or anti-parallel edges. If case of any doubts, call
        ``hynet.utilities.graph.eliminate_parallel_edges`` before assigning.
    roots : numpy.ndarray[.hynet_int_]
        Root nodes of the undirected graph defined by ``edges`` that specifies
        the sparsity pattern. At these root nodes the absolute angle of the
        optimal ``v`` is set to zero. (Note that, due to the quadratic form in
        ``v``, the optimal objective value is invariant w.r.t. an absolute phase
        shift within a connected component of the sparsity-defining graph.)
    normalization : QCQPPoint
        The attributes ``v``, ``f``, ``s``, and ``z`` specify the scaling of
        the corresponding optimization variable. This scaling is considered
        in the creation of the QCQP result, where the optimal value as well
        as the dual variables of the box constraints are adjusted accordingly.
    initial_point : QCQPPoint or None
        Initial point for the solver or None if omitted. (The initial point
        is typically only utilized in solvers for the nonconvex QCQP, but
        ignored in SDR and SOCR solvers.)
    use_buffered_vectorization : bool
        If True (default), the constraint vectorization is buffered to avoid
        its repeated computation. This QCQP object supports a vectorized
        representation, which is primarily intended for solver interfaces
        of relaxations, cf. ``get_vectorization``. As the vectorization of the
        constraints is computationally demanding, the result is buffered and
        reused if called repeatedly. If the *constraints are modified* between
        the calls, the buffering must be deactivated to update the
        vectorization.
    """

    def __init__(self, obj_func, eq_crt, ineq_crt, lb, ub, edges, roots,
                 normalization=None, initial_point=None,
                 use_buffered_vectorization=True):
        self.obj_func = obj_func
        self.eq_crt = eq_crt
        self.ineq_crt = ineq_crt
        self.lb = lb
        self.ub = ub
        self.edges = edges
        self.roots = roots
        if normalization is not None:
            self.normalization = normalization
        else:
            self.normalization = QCQPPoint(v=1.0, f=1.0, s=1.0, z=1.0)
        self.initial_point = initial_point
        self.use_buffered_vectorization = use_buffered_vectorization
        self._vectorization_buffer = None

    @property
    def dim_v(self):
        """Return the dimension of the ambient space of ``v``."""
        return len(self.lb.v)

    @property
    def dim_f(self):
        """Return the dimension of the ambient space of ``f``."""
        return len(self.lb.f)

    @property
    def dim_s(self):
        """Return the dimension of the ambient space of ``s``."""
        return len(self.lb.s)

    @property
    def dim_z(self):
        """Return the dimension of the ambient space of ``z``."""
        return len(self.lb.z)

    @property
    def num_edges(self):
        """Return the number of edges in the sparsity-defining graph."""
        return len(self.edges[0])

    def __repr__(self):
        return ('QCQP(v:C^{0}, f:R^{1}, s:R^{2}, z:R^{3} | '
                '#EQ={4}, #INEQ={5}, #edges={6}, #roots={7})'
                ).format(self.dim_v, self.dim_f, self.dim_s, self.dim_z,
                         len(self.eq_crt), len(self.ineq_crt),
                         self.num_edges, len(self.roots))

    def _mtx2vec(self, A):
        """
        Return a vector "``a^T``" such that ``tr(A*V) = a^T v_tld``.

        Therein, ``v_tld`` is the vectorization of ``V`` as used in the
        vectorized problem. The matrix ``A`` is assumed to be Hermitian and
        to exhibit the sparsity pattern defined by the edges.

        **Remark:** For a Hermitian matrix ``A`` and ``V``, the inner product
        ``tr(A*V)`` is the sum of the product of their diagonal elements plus
        two times the sum of the product of the real and imaginary part of the
        off-diagonal elements. In this vectorization, the factor of two is
        included in the vectorization of the coefficient matrix ``A``, while
        the vectorization ``v_tld`` of the optimization variable ``V`` only
        comprises the stacking of the respective variables. (Note that, for
        symmetry reasons, the off-diagonal elements are typically scaled by
        ``sqrt(2)``, but we deviate from that due to implementation reasons.)
        """
        if self.num_edges:
            A = A.tocsr()  # Conversion to CSR for indexing
            A_offd = np.asarray(A[self.edges[0], self.edges[1]])[0]
            return hynet_sparse_(np.concatenate((A.diagonal().real,
                                                 2*A_offd.real,
                                                 2*A_offd.imag)))
        return hynet_sparse_(A.diagonal().real)

    def _vectorize_constraints(self, constraints):
        """
        Return ``(A, b)`` that represent the constraints in vectorized form.

        This method reformulates the given constraints of the form
        ::

            v^H C_i v + c_i^T f + a_i^T s + r_i^T z <=/== b_i

        into
        ::

            [ _mtx2vec(C_i), c_i^T, a_i^T, r_i^T ] x <=/== b_i

        with x as defined in the documentation of ``get_vectorization`` and
        _mtx2vec as defined in the member function of the same name of this
        class. Finally, the vectorized constraints are stacked, i.e.,
        ::

            [ _mtx2vec(C_1), c_1^T, a_1^T, r_1^T ]         [b_1]
            [                                    ]         [   ]
            [ ...                                ] x <=/== [...]
            [                                    ]         [   ]
            [ _mtx2vec(C_N), c_N^T, a_N^T, r_N^T ]         [b_N]

            \____________________________________/         \___/
                             A                     x <=/==   b

        Parameters
        ----------
        constraints : Iterable[Constraint]
            Constraint that shall be vectorized.

        Returns
        -------
        A : .hynet_sparse_
        b : numpy.ndarray[.hynet_float_]
        """
        timer = Timer()
        ofs_v_real = self.dim_v
        ofs_v_imag = ofs_v_real + self.num_edges
        ofs_c = ofs_v_imag + self.num_edges
        ofs_a = ofs_c + self.dim_f
        ofs_r = ofs_a + self.dim_s
        dim_x = ofs_r + self.dim_z

        aux = _AuxiliaryData(ofs_v_real=ofs_v_real,
                             ofs_v_imag=ofs_v_imag,
                             ofs_c=ofs_c,
                             ofs_a=ofs_a,
                             ofs_r=ofs_r,
                             dim_x=dim_x,
                             e_src=self.edges[0],
                             e_dst=self.edges[1])

        data = zip(range(len(constraints)),
                   constraints,
                   [aux] * len(constraints))

        result_list = workers.starmap(_vectorize_constraint, data)

        list_idx_row, list_idx_col, list_data = zip(*result_list)
        A = create_sparse_matrix(np.concatenate(list_idx_row),
                                 np.concatenate(list_idx_col),
                                 np.concatenate(list_data),
                                 len(constraints), dim_x, dtype=hynet_float_)
        b = np.array([crt.b for crt in constraints], dtype=hynet_float_)

        _log.debug("Constraint vectorization ({:.3f} sec.)"
                   .format(timer.total()))
        return A, b

    def get_constraint_vectorization(self):
        """
        Return a vectorized representation of the constraints.

        This class supports a vectorized representation of the QCQP, see
        ``get_vectorization`` for more details. This method returns the
        parameters for the vectorization of the equality and inequality
        constraints, i.e., ``Ax == b`` and ``Bx <= d``.

        See Also
        --------
        get_vectorization

        Returns
        -------
        A, b, B, d : tuple
            Vectorization of the equality and inequality constraints.
        """
        if self.use_buffered_vectorization and \
                self._vectorization_buffer is not None:
            _log.debug("Constraint vectorization ~ Using buffered result.")
            return self._vectorization_buffer

        A, b = self._vectorize_constraints(self.eq_crt)
        B, d = self._vectorize_constraints(self.ineq_crt)
        self._vectorization_buffer = (A, b, B, d)  # Buffer the vectorization

        return A, b, B, d

    def get_vectorization(self):
        """
        Return a vectorized representation of this QCQP::

            min   c^T x   s.t.   Ax == b,  Bx <= d,  x_lb <= x <= x_ub
             x

        This representation is primarily intended for solver interfaces to
        *relaxations* of the QCQP and based on a reformulation in three steps,
        i.e.,

        1)  The complex-valued optimization variable ``v`` is replaced by a
            matrix ``V``::

                v^H C v = tr(v^H C v) = tr(Cvv^H) = tr(CV) with V = vv^H

            For equivalence to the original problem, ``V`` must be Hermitian
            (``V = V^H``), positive semidefinite (``V >= 0``) and have rank 1
            (``rank(V) = 1``). **Note that this vectorization does not take
            care of this, it only reformulates the problem in in terms of
            ``V``.** (For example, the semidefinite relaxation will include
            the psd constraint but omit the rank constraint.)

        2)  The sparsity of the constraint matrices is utilized to reduce the
            number of effective variables and vectorize the matrices, i.e.,
            ::

                tr(CV) = c_vec^T v_tld

            where ``c_vec = self._mtx2vec(C)`` and
            ::

                v_tld = [V_{1,1}                       ^
                         ...                           | N = self.dim_v
                         V_{N,N}                       v
                         real(V_{e_src(1),e_dst(1)}    ^
                         ...                           | K = self.num_edges
                         real(V_{e_src(K),e_dst(K)}    v
                         imag(V_{e_src(1),e_dst(1)}    ^
                         ...                           | K = self.num_edges
                         imag(V_{e_src(K),e_dst(K)}]   v

        3)  All optimization variables are stacked, i.e.,
            ::

                x = [v_tld^T, f^T, s^T, z^T]^T

            and the objective as well as the equality, inequality, and box
            constraints are adapted accordingly.

        **Caution:** This representation is **only reasonable with additional
        constraints** on the ``v_tld``-part of ``x``. For example, psd
        constraints on principal submatrices for the SOC relaxation or a psd
        constraint on the "reassembled" matrix ``V`` for the SDR.

        Returns
        -------
        (c, A, b, B, d, x_lb, x_ub) : tuple
            Parameters of the vectorized QCQP representation.
        """
        timer = Timer()

        c = hynet_sparse_(scipy.sparse.vstack(
            (
                self._mtx2vec(self.obj_func.C).transpose(),
                self.obj_func.c,
                self.obj_func.a,
                self.obj_func.r
            )))

        A, b, B, d = self.get_constraint_vectorization()

        v_bnd_od = np.multiply(self.ub.v[self.edges[0]],
                               self.ub.v[self.edges[1]])
        x_lb = np.concatenate((np.square(self.lb.v),
                               -v_bnd_od,
                               -v_bnd_od,
                               self.lb.f,
                               self.lb.s,
                               self.lb.z
                               ))
        x_ub = np.concatenate((np.square(self.ub.v),
                               v_bnd_od,
                               v_bnd_od,
                               self.ub.f,
                               self.ub.s,
                               self.ub.z
                               ))

        _log.debug("QCQP vectorization ({:.3f} sec.)".format(timer.total()))
        return c, A, b, B, d, x_lb, x_ub

    def split_vectorization_optimizer(self, x):
        """
        Return ``(V, f, s, z)`` for the optimizer ``x`` of the vectorized problem.

        The partitioning in ``x`` is assumed as defined in ``get_vectorization``.
        The vectorized representation ``v_tld`` of ``V`` is retracted, i.e.,
        ``V`` is populated with the diagonal and off-diagonal elements
        specified by ``v_tld``.
        """
        (e_src, e_dst) = self.edges
        N = self.dim_v
        K = self.num_edges
        ofs = 0

        V_diag = x[ofs:(ofs + N)]
        ofs += N
        V_offd = x[ofs:(ofs + K)] + 1j * x[(ofs + K):(ofs + 2*K)]
        ofs += 2*K

        f = x[ofs:(ofs + self.dim_f)]
        ofs += self.dim_f

        s = x[ofs:(ofs + self.dim_s)]
        ofs += self.dim_s

        z = x[ofs:(ofs + self.dim_z)]
        ofs += self.dim_z

        V = create_sparse_diag_matrix(V_diag) \
          + create_sparse_matrix(e_src, e_dst, V_offd, N, N) \
          + create_sparse_matrix(e_dst, e_src, V_offd.conj(), N, N)

        return hynet_sparse_(V), f, s, z

    def split_vectorization_bound_dual(self, dual_var):
        """
        Return the dual variable of a bound on ``x`` as a QCQP point.

        Splits the dual variable of a bound on the optimization variable ``x``
        of the vectorized problem (``x >= x_lb`` or ``x <= x_ub``) into the
        dual variables of the respective bound on ``f``, ``s``, and ``z`` and
        returns them as a QCQP point (with ``v`` set to ``None`` as the
        respective magnitude bounds are optional).

        Parameters
        ----------
        dual_var : numpy.ndarray[.hynet_float_]
            Dual variable of a bound on ``x`` of the vectorized problem.

        Returns
        -------
        point : QCQPPoint
            QCQP point with the dual variables of the respective bounds on
            ``|v|``, ``f``, ``s``, and ``z``.
        """
        (_, dv_f, dv_s, dv_z) = self.split_vectorization_optimizer(dual_var)
        return QCQPPoint(v=None, f=dv_f, s=dv_s, z=dv_z)


class _AuxiliaryData(namedtuple('_AuxiliaryData',
                                ['ofs_v_real', 'ofs_v_imag', 'ofs_c', 'ofs_a',
                                 'ofs_r', 'dim_x', 'e_src', 'e_dst'])):
    """
    Auxiliary data for the parallelization of constraint vectorization.
    """
    pass


def _vectorize_constraint(i, crt, aux):
    """
    Return the sparse data for the vectorized constraint at position ``i``.

    See Also
    --------
    QCQP._vectorize_constraints
    """
    # This vectorization was a bottleneck of the code and I had to sacrifice
    # readability for performance. I apologize for any inconvenience caused ;)
    list_idx_row = []
    list_idx_col = []
    list_data = []

    if crt.C is not None:
        # The following code corresponds to the operation of QCQP._mtx2vec
        C_diag = crt.C.diagonal().real
        idx_col = C_diag.nonzero()[0]
        if idx_col.size:
            list_idx_row.append(i + np.zeros(len(idx_col), dtype=hynet_int_))
            list_idx_col.append(idx_col)
            list_data.append(C_diag[idx_col])
        if aux.e_src.size:
            C_offd = np.asarray(crt.C.tocsr()[aux.e_src, aux.e_dst])[0]
            idx_col = C_offd.nonzero()[0]
            if idx_col.size:
                idx_row = i + np.zeros(len(idx_col), dtype=hynet_int_)
                data = C_offd[idx_col]
                list_idx_row.append(idx_row)
                list_idx_col.append(idx_col + aux.ofs_v_real)
                list_data.append(2*data.real)
                list_idx_row.append(idx_row)
                list_idx_col.append(idx_col + aux.ofs_v_imag)
                list_data.append(2*data.imag)
    if crt.c is not None:
        # REMARK: This code includes the transposition by exchanging
        # the row and column index and it utilizes the fact that the
        # column indices are always zero to avoid an additional vector
        # allocation. The same applies to crt.a and crt.r.
        if crt.c.row.size:
            list_idx_row.append(crt.c.col + i)
            list_idx_col.append(crt.c.row + aux.ofs_c)
            list_data.append(crt.c.data)
    if crt.a is not None:
        if crt.a.row.size:
            list_idx_row.append(crt.a.col + i)
            list_idx_col.append(crt.a.row + aux.ofs_a)
            list_data.append(crt.a.data)
    if crt.r is not None:
        if crt.r.row.size:
            list_idx_row.append(crt.r.col + i)
            list_idx_col.append(crt.r.row + aux.ofs_r)
            list_data.append(crt.r.data)

    if list_idx_row:
        row = np.concatenate(list_idx_row)
        col = np.concatenate(list_idx_col)
        data = np.concatenate(list_data)
    else:
        row = col = np.ndarray(0, dtype=hynet_int_)
        data = np.ndarray(0, dtype=hynet_float_)

    return row, col, data



#pylint: disable=no-member,logging-format-interpolation
"""
*hynet* optimization server for distributed computation.
"""

import logging
import subprocess
import getpass
from multiprocessing import Queue, Value, Process
from multiprocessing.managers import SyncManager
from time import sleep

import numpy as np

import hynet.config as config
from hynet.scenario.representation import Scenario
from hynet.model.steady_state import SystemModel
from hynet.qcqp.problem import QCQP
from hynet.distributed.client import start_optimization_client

_log = logging.getLogger(__name__)


def start_optimization_server(port=None, authkey=None, num_local_workers=0):
    """
    Create, start, and return a *hynet* optimization server.

    Parameters
    ----------
    port : int, optional
        TCP port on which the *hynet* optimization server shall be running.
    authkey : str, optional
        Authentication key that must be presented by *hynet* optimization
        clients to connect to the server.
    num_local_workers : int, optional
        Number of worker processes that shall run on the local machine
        (zero by default). Note that the calculation of optimization jobs
        (see ``calc_jobs``) is blocked until at least one *hynet*
        optimization client is connected to the server. By adding local
        workers, jobs are also processed on the server machine, which
        additionally prevents a block of the job processing. If more than one
        local worker is started, it is recommended to disable the internal
        parallel processing, see ``parallelize`` in ``hynet.config``.

    Returns
    -------
    server : OptimizationServer
        The *hynet* optimization server.
    """
    if port is None:
        port = config.DISTRIBUTED['default_port']
    if authkey is None:
        authkey = config.DISTRIBUTED['default_authkey']

    server = OptimizationServer(port, authkey)

    if num_local_workers > 0:
        local_client = Process(target=start_optimization_client,
                               args=(config.DISTRIBUTED['local_ip'],),
                               kwargs={
                                   'port': port,
                                   'authkey': authkey,
                                   'num_workers': num_local_workers,
                                   'verbose': False
                               })
        local_client.start()

    return server


class OptimizationJob:
    """
    Represents a *hynet* optimization job.

    A *hynet* optimization job may be an OPF problem, specified via the
    scenario data or system model, or a QCQP problem. Furthermore, the solver
    or solver type for its solution may be specified.

    Parameters
    ----------
    problem : Scenario or SystemModel or QCQP
        Problem specification.
    solver : SolverInterface, optional
        Solver for the provided problem; the default selects an appropriate
        solver of those available. For QCQPs, the solver must be specified
        explicitly. Please make sure that the selected solver is installed on
        all client machines.
    solver_type : SolverType, optional
        Optional solver type for the problem. If passed, it restricts the
        automatic solver selection to this type. It is ignored if ``solver``
        is not ``None``.
    """

    def __init__(self, problem, solver=None, solver_type=None):
        if not isinstance(problem, (Scenario, SystemModel, QCQP)):
            raise ValueError("The provided problem specification is invalid.")

        if isinstance(problem, QCQP) and solver is None:
            raise ValueError("For QCQPs, the solver must be provided.")

        self.problem = problem
        self.solver = solver
        self.solver_type = solver_type
        self.id = None  # Used internally to sort the job results


class OptimizationServer:
    """
    *hynet* optimization server for distributed computation.

    This server manages the distributed computation of a set of *hynet*
    optimization problems (OPF or QCQPs) on *hynet* optimization clients.
    """

    def __init__(self, port, authkey):
        """
        Create a *hynet* optimization server.
        """
        self._port = port
        self._authkey = authkey
        self._manager = _create_server_manager(port, authkey.encode('utf-8'))
        self._job_queue = self._manager.get_job_queue()
        self._result_queue = self._manager.get_result_queue()

    def calc_jobs(self, job_list, solver=None, show_progress=True):
        """
        Calculate the list of *hynet* optimization jobs and return the results.

        The provided list of jobs is processed by distributing them to the
        connected *hynet* optimization clients, collecting the results, and
        returning an array of results that corresponds with the provided array
        of jobs. Note that if there are no clients connected, this method will
        wait until a client is connected to process the jobs.

        Parameters
        ----------
        job_list : array-like
            List of *hynet* optimization jobs (``OptimizationJob``) or problem
            specifications (``Scenario``, ``SystemModel``, or ``QCQP``).
        solver : SolverInterface, optional
            If provided, this solver is used for problem specifications
            (``Scenario``, ``SystemModel``, or ``QCQP``). It is ignored for job
            specifications (``OptimizationJob``).
        show_progress : bool, optional
            If True (default), the progress is reported to the standard output.

        Returns
        -------
        results : numpy.ndarray
            Array containing the optimization results.
        """

        for i, job in enumerate(job_list):
            if not isinstance(job, OptimizationJob):
                job = OptimizationJob(job, solver=solver)
            job.id = i + 1
            self._job_queue.put(job)

        num_results = Value('I', 0, lock=False)  # Unsigned int shared mem. var.
        results = np.empty(len(job_list), dtype=object)

        if show_progress:
            progress_bar = Process(target=_progress_bar,
                                   args=(num_results, len(job_list)))
            progress_bar.start()

        while num_results.value < len(job_list):
            result_dict = self._result_queue.get()
            for (job_id, result) in result_dict.items():
                results[job_id - 1] = result
            num_results.value += len(result_dict)

        if show_progress:  # Wait for the progress bar to finish
            progress_bar.join()

        return results

    def shutdown(self):
        """Stop the *hynet* optimization server and all connected clients."""
        self._manager.shutdown()

    def start_clients(self, client_list, server_ip, ssh_user=None,
                      ssh_port=None, num_workers=None, log_file=None):
        """
        Automated start of *hynet* optimization clients.

        This method provides an automatic start of *hynet* optimization clients
        via SSH if the server can connect to the clients via ``ssh [client]``
        (e.g. by configuring SSH keys; please be aware of the related aspects
        of system security). *hynet* must be properly installed on all client
        machines.

        This function uses SSH to run the *hynet* package with the sub-command
        ``client`` and corresponding command line arguments (``python -m hynet
        client ...``) on every client machine. To customize the SSH and Python
        command, see ``hynet.config``.

        Parameters
        ----------
        client_list : array-like
            List of strings containing the host names or IP addresses of the
            client machines.
        server_ip : str
            IP address the *hynet* optimization server.
        ssh_user : str, optional
            The user name for the SSH login on the client machines. By default,
            this is set to the current user name (``getpass.getuser()``).
        ssh_port : int, optional
            Port on which SSH is running on the client machines.
        num_workers : int, optional
            Number of worker processes that should run in parallel on every
            client machine.
        log_file : str, optional
            Log file on the client machines to capture the output.
        """
        if ssh_user is None:
            ssh_user = getpass.getuser()

        command_pre = config.DISTRIBUTED['ssh_command'] + " -f "
        if ssh_port is not None:
            command_pre += "-p {0:d} ".format(ssh_port)
        command_pre += ssh_user + "@"

        command_post = ' "' + config.DISTRIBUTED['python_command']
        command_post += " -m hynet client "
        command_post += server_ip
        command_post += " -p {0:d}".format(self._port)
        command_post += " -a " + self._authkey
        if num_workers is not None:
            command_post += " -n {0:d}".format(num_workers)
        if log_file is not None:
            command_post += " &> " + log_file
        command_post += '"'

        for client in client_list:
            try:
                subprocess.run(command_pre + client + command_post,
                               shell=True, check=True)
            except subprocess.CalledProcessError as exception:
                _log.error("Failed to start client on '{0}': {1}"
                           .format(client, str(exception)))


def _create_server_manager(port, authkey):
    """
    Return a (started) manager for a *hynet* optimization server.

    Parameters
    ----------
    port : int
        TCP port on which the *hynet* optimization server shall be running.
    authkey : str
        Authentication key that must be presented by *hynet* optimization
        clients to connect to the server.

    Returns
    -------
    manager : ServerManager
        Manager object for a *hynet* optimization server.
    """
    job_queue = Queue()
    result_queue = Queue()

    class ServerManager(SyncManager):
        """This class manages the synchronization of the queues."""
        pass

    ServerManager.register('get_job_queue', callable=lambda: job_queue)
    ServerManager.register('get_result_queue', callable=lambda: result_queue)

    manager = ServerManager(address=('', port), authkey=authkey)
    manager.start()
    _log.info('Started server on port {0}'.format(port))
    return manager


def _progress_bar(counter, num_total):
    """
    Show a progress bar on the standard output.

    Parameters
    ----------
    counter : multiprocessing.Value
        Counter for the number of processed jobs as a shared memory object.
    num_total : int
        Total number of jobs.
    """
    while True:
        num_done = counter.value
        percentage_done = (num_done/num_total)*100

        len_done = int(percentage_done/2)
        progress_bar = "\r[" + "="*len_done
        if percentage_done < 100:
            progress_bar += ">"
            progress_bar += "-"*(49 - len_done)
        progress_bar += "] "
        progress_bar += "{0:d}% ".format(int(percentage_done))
        progress_bar += "{0:d}/{1:d}".format(num_done, num_total)
        print(progress_bar, end='', flush=True)

        if num_done == num_total:
            print("")
            return
        sleep(0.25)



"""
Settings view to edit parameters in the capability region visualizer.
"""

import logging
from re import compile as compile_regex

import tkinter as tk
from tkinter import ttk

from hynet.scenario.capability import HalfSpace, CapRegion
from hynet.visual.capability.utilities import Orientation

_log = logging.getLogger(__name__)

DIGITS = 5


class SettingsView(tk.ttk.Frame):  # pylint: disable=too-many-ancestors
    """
    Settings view to edit parameters in the capability region visualizer.
    """
    def __init__(self, master=None, cap_region=CapRegion()):
        super(SettingsView, self).__init__(master)
        self.error_callback = None
        self.container = tk.Frame(master, padx=0, pady=0, bg="white")
        self.capability_region = cap_region
        power_tuple = (cap_region.p_min, cap_region.p_max)
        reactive_tuple = (cap_region.q_min, cap_region.q_max)
        # active power view (i.e., first row)
        self.active_power = ValueView(self.container,
                                      'Active power:',
                                      left_value_name='Minimum:',
                                      right_value_name='Maximum:',
                                      initial_value_tuple=power_tuple)
        self.active_power.container.grid(row=0, column=0, sticky=[tk.W, tk.E])

        # reactive power view (i.e., second row)
        self.reactive_power = ValueView(self.container,
                                        'Reactive power:',
                                        left_value_name='Minimum:',
                                        right_value_name='Maximum:',
                                        initial_value_tuple=reactive_tuple)
        self.reactive_power.container.grid(row=1, column=0,
                                           sticky=[tk.W, tk.E])

        # halfplane views (i.e., third to sixth row)
        self.rt_half_space = ValueView(self.container,
                                       'Right-top half-space:',
                                       initial_value_tuple=cap_region.rt,
                                       checkbox_state=cap_region.rt is not None)
        self.rt_half_space.container.grid(row=2, column=0,
                                          sticky=[tk.W, tk.E])

        self.rb_half_space = ValueView(self.container,
                                       'Right-bottom half-space:',
                                       initial_value_tuple=cap_region.rb,
                                       checkbox_state=cap_region.rb is not None)
        self.rb_half_space.container.grid(row=3, column=0,
                                          sticky=[tk.W, tk.E])

        self.lt_half_space = ValueView(self.container,
                                       'Left-top half-space:',
                                       initial_value_tuple=cap_region.lt,
                                       checkbox_state=cap_region.lt is not None)
        self.lt_half_space.container.grid(row=4, column=0,
                                          sticky=[tk.W, tk.E])

        self.lb_half_space = ValueView(self.container,
                                       'Left-bottom half-space:',
                                       initial_value_tuple=cap_region.lb,
                                       checkbox_state=cap_region.lb is not None)
        self.lb_half_space.container.grid(row=5, column=0,
                                          sticky=[tk.W, tk.E])

        self.power_factor = PowerFactorView(self.container)
        self.power_factor.container.grid(row=6, column=0, sticky=[tk.W, tk.E])
        spacer = tk.Frame(self.container, bg="white")
        spacer.grid(row=7, column=0, sticky=[tk.W, tk.E, tk.S, tk.N])

        self.rowconfigure(7, weight=1)

        # setup all error_callbacks
        for value_view in [self.lt_half_space,
                           self.lb_half_space,
                           self.rt_half_space,
                           self.rb_half_space, self.active_power,
                           self.reactive_power]:
            value_view.error_callback = self._proxy_error_callback

    def setup_halfspace_callbacks(self, update_view_callback):
        """
        Set the callback that updates the view with changes of the half-spaces.

        Parameters
        ----------
        update_view_callback: function
            This function is called when the checkbox is (de)activated and
            should update/reload the view of the capability region.
        """
        def _create_callback(value_view, cap_region, orientation_tuple):
            def _callback(value):
                if value:
                    slope = value_view.right_entry.get_value()
                    offset = value_view.left_entry.get_value()
                    new_halfspace = HalfSpace(slope=slope, offset=offset)
                else:
                    new_halfspace = None
                active_power, reactive_power = orientation_tuple
                raise_error = False
                try:
                    if active_power is Orientation.RIGHT and \
                            reactive_power is Orientation.TOP:
                        cap_region.rt = new_halfspace
                    elif active_power is Orientation.RIGHT and \
                            reactive_power is Orientation.BOTTOM:
                        cap_region.rb = new_halfspace
                    elif active_power is Orientation.LEFT and \
                            reactive_power is Orientation.TOP:
                        cap_region.lt = new_halfspace
                    elif active_power is Orientation.LEFT and \
                            reactive_power is Orientation.BOTTOM:
                        cap_region.lb = new_halfspace
                    else:
                        raise_error = True
                except ValueError as halfspace_invalid_error:
                    # halfspace is invalid and thus capregion.(l|r)(t|b) = x
                    # raises an error => load the values of the capregion into
                    # the visual again / reset the user's input
                    self._proxy_error_callback(
                        halfspace_invalid_error,
                        undo_callback=self.update_from_capability_region)
                if raise_error:
                    raise ValueError(
                        f"Unhandled case: ({active_power}, {reactive_power}")
                update_view_callback()

            return _callback

        self.lt_half_space.callback = _create_callback(
            self.lt_half_space, self.capability_region,
            (Orientation.LEFT, Orientation.TOP))
        self.lb_half_space.callback = _create_callback(
            self.lb_half_space, self.capability_region,
            (Orientation.LEFT, Orientation.BOTTOM))
        self.rt_half_space.callback = _create_callback(
            self.rt_half_space, self.capability_region,
            (Orientation.RIGHT, Orientation.TOP))
        self.rb_half_space.callback = _create_callback(
            self.rb_half_space, self.capability_region,
            (Orientation.RIGHT, Orientation.BOTTOM))

    def deactivate_all_halfspaces(self):
        """
        Deactivate all checkboxes and, therewith, remove all half-spaces.
        """
        self.lt_half_space.checkbox_value.set(False)
        self.lb_half_space.checkbox_value.set(False)
        self.rt_half_space.checkbox_value.set(False)
        self.rb_half_space.checkbox_value.set(False)

    def update_from_capability_region(self):
        """
        Update the views with the (possibly) changed capability region.
        """
        self.lt_half_space.update_from_capability_region(
            self.capability_region.lt)
        self.lb_half_space.update_from_capability_region(
            self.capability_region.lb)
        self.rt_half_space.update_from_capability_region(
            self.capability_region.rt)
        self.rb_half_space.update_from_capability_region(
            self.capability_region.rb)

    def _proxy_error_callback(self, message, undo_callback=None):
        if self.error_callback:
            # pylint: disable=not-callable
            self.error_callback(message, undo_callback)
        else:
            raise ValueError("Received an error call, "
                             "but no callback is configured.")


class PowerFactorView(tk.ttk.Frame):  # pylint: disable=too-many-ancestors
    """View for setting the power factor limit."""
    def __init__(self, master=None):
        super().__init__(master)
        self.container = tk.Frame(master, padx=0, pady=0, bg="white")
        self._title = tk.Label(self.container,
                               text="Set power factor:",
                               anchor=tk.SW,
                               pady=1,
                               bg="white")
        self._title.grid(row=0, column=0, columnspan=2, sticky=[tk.W, tk.E])
        self._callback = lambda x: x
        self._value = tk.StringVar()
        self._entry = tk.Entry(self.container,
                               textvariable=self._value,
                               bg="white",
                               width=10)
        self._entry.grid(row=1, column=0, padx=(25, 5), sticky=(tk.W, tk.E))

        self._trigger = tk.Button(self.container,
                                  text="Apply",
                                  padx=5,
                                  command=self._proxy_callback,
                                  bg="white",
                                  width=10)
        self._trigger.grid(row=1, column=1, padx=0, pady=0,
                           sticky=(tk.W, tk.E, tk.S, tk.N))

    def _proxy_callback(self):
        value = self._value.get()
        if not value:
            return
        try:
            self._callback(float(value))
        except ValueError:
            _log.warning(value + " is not a float!")
        self._entry.delete(0, tk.END)

    def set_callback(self, callback_function):
        """
        Set the callback for updating the power factor limit.

        Parameters
        ----------
        callback_function: function
            Function that takes the power factor as an argument and updates
            the capability region.
        """
        self._callback = callback_function


class ValueView(tk.ttk.Frame):  # pylint: disable=too-many-ancestors
    """
    Container view for two descriptive entries.

    Furthermore, the value view handles the update of the associated half-space
    (if provided) as well as the error and update callbacks.
    """
    callback = None

    def __init__(self, master=None, name='Unnamed ValueView',
                 right_value_name='Slope:', left_value_name='Offset:',
                 initial_value_tuple=(0, 0), checkbox_state=None):
        super(ValueView, self).__init__(master)
        self.error_callback = None
        if initial_value_tuple:
            left, right = initial_value_tuple
        else:
            left, right = (0.0, 0.0)
        self.container = tk.Frame(master, padx=5, pady=5, bg='white')

        if checkbox_state is None:
            self._title = tk.Label(self.container,
                                   anchor=tk.SW,
                                   text=name,
                                   bg="white")
        else:
            self.checkbox_value = tk.BooleanVar()
            self.checkbox_value.set(checkbox_state)
            self._title = tk.Checkbutton(self.container,
                                         text=name,
                                         variable=self.checkbox_value,
                                         command=self._proxy_callback,
                                         anchor=tk.SW,
                                         bg="white")

        self._title.grid(row=0, column=0, columnspan=2,
                         padx=0, pady=0, sticky=[tk.W, tk.E])

        self.left_entry = DescriptiveEntry(self.container, left_value_name,
                                           alignment=tk.VERTICAL, value=left)
        self.left_entry.container.grid(row=1, column=0, padx=(20, 5), pady=0,
                                       sticky=[tk.W, tk.E])

        self.right_entry = DescriptiveEntry(self.container, right_value_name,
                                            alignment=tk.VERTICAL,
                                            value=right)
        self.right_entry.container.grid(row=1, column=1, padx=(0, 5), pady=0,
                                        sticky=[tk.W, tk.E])

        self.left_entry.error_callback = self._proxy_error_callback
        self.right_entry.error_callback = self._proxy_error_callback

    def _proxy_error_callback(self, message, undo_callback=None):
        if self.error_callback:
            # pylint: disable=not-callable
            self.error_callback(message, undo_callback)
        else:
            raise ValueError("Received an error call, "
                             "but no callback is configured.")

    def _proxy_callback(self):
        if self.callback:
            try:
                # pylint: disable=not-callable
                self.callback(self.checkbox_value.get())
            except ValueError as error:
                self._proxy_error_callback(str(error))
        else:
            raise ValueError("Received a call, but no callback is configured.")

    def update_from_capability_region(self, halfspace):
        """
        Updates this view based on the provided half-space specification.

        Parameters
        ----------
        halfspace: HalfSpace
            The half-space associated with this value view.
        """
        if halfspace:
            left, right = halfspace
            self.left_entry.reset_to_normal_state(round(left, DIGITS))
            self.right_entry.reset_to_normal_state(round(right, DIGITS))
            self.checkbox_value.set(True)
        else:
            self.checkbox_value.set(False)


class DescriptiveEntry(ttk.Frame):  # pylint: disable=too-many-ancestors
    """Container that bundles a label with an input field."""
    VALIDATION_REGEX = compile_regex(r'^[+-]?([0-9]*[.])?[0-9]+$')

    def __init__(self, master=None, name='Label', alignment=tk.HORIZONTAL,
                 value=0.0):
        super(DescriptiveEntry, self).__init__(master)

        self.container = tk.Frame(master, padx=0, pady=0, bg="white")
        self._label = tk.Label(self.container,
                               text=name,
                               anchor=tk.SW,
                               bg="white",
                               width=10)
        self._label.grid(row=0, column=0, padx=0, pady=0)
        self._value = tk.StringVar(value=value)
        self._last_value = value
        self._validator = (self.register(self._validation_callback), '%P')

        self._entry = tk.Entry(self.container,
                               textvariable=self._value,
                               validate="focus",
                               validatecommand=self._validator,
                               bg="white",
                               width=10)

        if alignment is tk.HORIZONTAL:
            self._entry.grid(row=0, column=1, padx=0, pady=0, sticky=tk.W)
        else:
            self._entry.grid(row=1, column=0, padx=0, pady=0, sticky=tk.W)

    def _validation_callback(self, new_text):
        if self.VALIDATION_REGEX.match(new_text):
            return True

        undo = self.error_state()
        message = f"{new_text} is not a valid number!"
        self._proxy_error_callback(message, undo_callback=undo)
        _log.warning(message)
        return False

    def error_state(self, old_value=None):
        """
        Set the foreground to red and return the callback to undo this change.

        Parameters
        ----------
        old_value: float
            If this value is set, then the value of the input component is set
            to this value when this callback is called.

        Returns
        -------
        function
            Callback to undo this change (set the foreground back to black).
        """
        self._entry.configure(foreground='red')
        return lambda: self.reset_to_normal_state(old_value)

    def reset_to_normal_state(self, old_value=None):
        """
        Reset the element to the normal state (black font on white background).

        Parameters
        ----------
        old_value: float
            If this value is set, then the value of the input component is set
            to this value when this callback is called.
        """
        self._entry.configure(foreground='black')
        if old_value:
            self._value.set(old_value)
            self._last_value = old_value

    def _proxy_error_callback(self, message, undo_callback=None):
        if self.error_callback:
            self.error_callback(message, undo_callback)
        else:
            raise ValueError("Received an error call, "
                             "but no callback is configured.")

    def set_callback(self, callback):
        """
        Set a callback that listens for value changes.

        Parameters
        ----------
        callback: function
            The function that should be called when a value has changed.
        """

        def _get_value(*_args):
            try:
                value = float(self._value.get())
                # the order is important as the callback is used to check for
                # conformance of the new value!
                callback(value)
                self._last_value = value
            except ValueError as error:
                undo = self.error_state(old_value=self._last_value)
                self._proxy_error_callback(str(error), undo_callback=undo)

        self._entry.bind("<FocusOut>", _get_value)

    def get_value(self):
        """
        Return the current value of the field.

        Returns
        -------
        float
        """
        try:
            value = float(self._value.get())
        except ValueError:
            value = self._last_value
        return value

    def set_value(self, value):
        """
        Set the value for the field.

        Parameters
        ----------
        value: float or str
        """
        self._value.set(value)



"""
*hynet* package configuration.

Parameters
----------
GENERAL : dict
    General settings.

    ``parallelize``: (``bool``)
        Enable or disable parallel processing in *hynet*. If True, certain
        procedures (e.g., the construction of constraint matrices for the OPF
        formulation) are parallelized if the system features more than one CPU.

OPF : dict
    Optimal power flow settings.

    ``copper_plate_init``: (``bool``)
        Enable or disable the computation of a copper plate based initial point
        for the solution of the nonconvex QCQP. By default, this is enabled if
        MOSEK is installed and disabled otherwise. For the currently supported
        solvers, only MOSEK provides sufficient performance such that the
        computation of the initial point improves overall performance, i.e.,
        that the reduced number of iterations of the QCQP solver outweighs the
        computational cost for the initial point.

DISTRIBUTED : dict
    Settings for distributed computation.

    ``default_port``: (``int``)
        Default optimization server TCP port.
    ``default_authkey``: (``str``)
        Default optimization server authentication key.
    ``default_num_workers``: (``int``)
        Default number of worker processes on an optimization client.
    ``local_ip``: (``str``)
        IP address of the local machine (for workers on the optimization
        server).
    ``ssh_command``: (``str``)
        Command to run SSH on the local machine.
    ``python_command``: (``str``)
        Command to run Python on client machines.
"""

from hynet.utilities.base import validate_mosek_license


GENERAL = {
    'parallelize': True
}

OPF = {
    'copper_plate_init': validate_mosek_license()
}

DISTRIBUTED = {
    'default_port': 1235,
    'default_authkey': 'hynet',
    'default_num_workers': 1,
    'local_ip': '127.0.0.1',
    'ssh_command': 'ssh',
    'python_command': 'python'
}
